import os
import itertools
import functools
import pprint
import re
import collections
import inspect
import shutil
import hashlib
import json

import pysam
import pandas as pd
import pyranges as pr

import handygenome
import handygenome.network as network
import handygenome.tools as tools
import handygenome.logutils as logutils
import handygenome.deco as deco
import handygenome.publicdb.ncbi as libncbi
import handygenome.publicdb.ncbi_cache as libncbicache
import handygenome.refgenome.refverfile_wogdf as refverfile_wogdf


# constants

MT_CHROMS = ('chrM', 'MT')

CHROM_PATSTRING_CHR = '(?P<chr>chr)'
CHROM_PATSTRING_NUMBER = '(?P<number>0*(?P<number_proper>[1-9][0-9]*))'
CHROM_PATSTRING_XY = '(?P<xy>[XY])'

PAT_NUMERIC_CHROM = re.compile(f'{CHROM_PATSTRING_CHR}?{CHROM_PATSTRING_NUMBER}', flags=re.I)
PAT_ASSEMBLED_CHROM = re.compile(f'{CHROM_PATSTRING_CHR}?({CHROM_PATSTRING_NUMBER}|{CHROM_PATSTRING_XY})', flags=re.I)
PAT_XY_CHROM = re.compile(f'{CHROM_PATSTRING_CHR}?{CHROM_PATSTRING_XY}', flags=re.I)

CHR1_NAME_PAT = re.compile(r'(chr)?0*1')

FASTA_URLS = {
    'GRCh37_1000genomes': 'http://ftp.1000genomes.ebi.ac.uk/vol1/ftp/technical/reference/human_g1k_v37.fasta.gz',
    'GRCh37_ucsc': 'https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/analysisSet/hg19.p13.plusMT.no_alt_analysis_set.fa.gz',
    'GRCh38_1000genomes': 'http://ftp.1000genomes.ebi.ac.uk/vol1/ftp/technical/reference/GRCh38_reference_genome/GRCh38_full_analysis_set_plus_decoy_hla.fa',
}

REFVER_DIR = os.path.join(handygenome.USERDATA_DIR, 'reference_files')
os.makedirs(REFVER_DIR, exist_ok=True)


############################
# chromosome name handlers #
############################

def normalize_chrom(chrom, strip_chr=False):
    """Examples:
        chr1 -> chr1
        chrX -> chrX
        chr001 -> chr1
        1 -> chr1
        X -> chrX
        MT -> chrM
        NC_000001.10 -> NC_000001.10
        chr19_gl000209_random -> chr19_gl000209_random
    """
    mat = PAT_ASSEMBLED_CHROM.fullmatch(chrom)
    if mat is None:
        if chrom in MT_CHROMS:
            if strip_chr:
                return 'MT'
            else:
                return 'chrM'
        else:  # e.g. chr19_gl000209_random, NC_000001.10
            return chrom
    else:  # e.g. chr1, 1, X, chrY, chr001
        if mat.group('xy') is None:
            if strip_chr:
                return mat.group('number_proper')
            else:
                return 'chr' + mat.group('number_proper')
        else:
            assert mat.group('number') is None
            if strip_chr:
                return mat.group('xy')
            else:
                return 'chr' + mat.group('xy')


def check_assembled_chrom(chrom):
    return PAT_ASSEMBLED_CHROM.fullmatch(chrom) is not None


def choose_assembled_chroms(chroms):
    return list(filter(check_assembled_chrom, chroms))


def check_xy_chrom(chrom):
    return PAT_XY_CHROM.fullmatch(chrom) is not None


##############
# exceptions #
##############

class UnknownRefverError(Exception):
    pass


class NoChr1NameError(Exception):
    pass


class AssemblyspecUnavailableError(Exception):
    pass


class NoXYError(Exception):
    pass


###############################
# RefverEntity and RefverInfo #
###############################

class RefverEntity(collections.UserDict):
    mandatory_keys = (
        'standard',
        'aliases',
        'RefSeq_refver',
        'fasta_path',
        'species',
    )
    optional_keys = (
        'repeatmasker_path',
        'edited_repeat_path',

        'raw_geneset_path',
        'edited_geneset_path',

        'raw_regulatory_path',
        'edited_regulatory_path',

        'dbsnp_path',
    )

    @classmethod
    def from_config(cls, config_data, calc_checksum=True):
        result = cls()
        result.calc_checksum = calc_checksum
        result.update(config_data)
        for key in cls.optional_keys:
            result.setdefault(key, None)
        result.sanitycheck()

        return result

    def sanitycheck(self):
        assert set(self.__class__.mandatory_keys).issubset(self.keys()), (
            f'Data must include these keys: {self.__class__.mandatory_keys}'
        )

        if self['RefSeq_refver'] is not None:
            if self['species'] is None:
                raise Exception(f'When "RefSeq_refver" is set, "species" must be also set.')
            if self['RefSeq_refver'] not in libncbi.GENOME_PATHS.get_available_refvers(self['species']):
                raise Exception(f'"RefSeq_refver" and "species" does not match according to RefSeq database.')
            if self['fasta_path'] is not None:
                chromlenhash_fasta = ChromLengthHash.from_fasta_path(self['fasta_path'])
                assemblyspec = AssemblySpec.from_RefSeq_refver(self['RefSeq_refver'], self['species'])
                chromlenhash_asmblspec = ChromLengthHash.from_assemblyspec(assemblyspec)
                if not chromlenhash_fasta.check_length_compatibility(chromlenhash_asmblspec):
                    raise Exception(
                        f'The fasta file and RefSeq assembly report do not '
                        f'coincide with the length of chromosomes.'
                    )
        ###
        if self['RefSeq_refver'] is None:
            if self['fasta_path'] is None:
                raise Exception(f'When "RefSeq_refver" is not set, "fasta_path" must be set.')

    def set_cached_properties(self):
        logutils.log(
            f'Initializing RefverEntity caches (refver={self["standard"]})', 
            level='info', add_locstring=False,
        )
        for key in (
            'fasta', 
            'chromdict', 
            'assemblyspec', 
            'chrom_converter', 
            'chromlenhash', 
            #'geneset_handle',
        ):
            try:
                _ = getattr(self, key)
            except AssemblyspecUnavailableError:
                pass

    ############################
    # data paths as attributes #
    ############################

    @property
    def datadir(self):
        datadir = os.path.join(REFVER_DIR, self['standard'])
        os.makedirs(datadir, exist_ok=True)
        return datadir

    @property
    def processed_fasta_path(self):
        return os.path.join(self.datadir, 'genome.fasta')

    @property
    def unedited_fasta_checksum_path(self):
        return os.path.join(self.datadir, 'raw_fasta_checksum')

    @property
    def unedited_fasta_wascustom_path(self):
        return os.path.join(self.datadir, 'raw_fasta_wascustom.json')

    @property
    def processed_geneset_path(self):
        return os.path.join(self.datadir, 'geneset.gff3.gz')

    ########################
    # important attributes #
    ########################

    @functools.cached_property
    def fasta(self):
        refverfile_wogdf.prepare_processed_fasta(
            standard_refver=self['standard'],
            edited_path=self.processed_fasta_path,
            custom_path=self['fasta_path'],
            RefSeq_refver=self['RefSeq_refver'],
            species=self['species'],
            chrom_converter=(None if self['RefSeq_refver'] is None else self.chrom_converter),
            unedited_file_checksum_path=self.unedited_fasta_checksum_path,
            unedited_file_wascustom_path=self.unedited_fasta_wascustom_path,
            calc_checksum=self.calc_checksum,
        )
        return pysam.FastaFile(self.processed_fasta_path)
    
    @functools.cached_property
    def chromdict(self):
        return ChromDict.from_fasta(self.fasta, refver=self['standard'])

    @functools.cached_property
    def assemblyspec(self):
        if self['RefSeq_refver'] is None:
            raise AssemblyspecUnavailableError(f'AssemblySpec not available for this refver.')
        else:
            return AssemblySpec.from_RefSeq_refver(self['RefSeq_refver'], self['species'])

    @functools.cached_property
    def chrom_converter(self):
        if self['fasta_path'] is None:
            return self.assemblyspec.get_chrom_converter_from_self()
        else:
            with pysam.FastaFile(self['fasta_path']) as fasta:
                contigs = fasta.references
                lengths = fasta.lengths
            return self.assemblyspec.get_chrom_converter(contigs, lengths)

    @functools.cached_property
    def chromlenhash(self):
        return ChromLengthHash.from_fasta(self.fasta)

    @functools.cached_property
    def geneset_handle(self):
        self.prepare_processed_geneset()
        return pysam.TabixFile(self.processed_geneset_path, parser=pysam.asGTF())


class RefverInfo:
    # init and helpers
    def __init__(self, initial_data, calc_checksum=True):
        self.refver_entities = list()
        for config_data in initial_data:
            refver_entity = RefverEntity.from_config(config_data, calc_checksum=calc_checksum)
            self.add_refver_entity(
                refver_entity, 
                reset_cache=False,
                set_refver_entity_cache=False,
            )

        #for refver_entity in self.refver_entities:
        #    refver_entity.set_cached_properties()

        self.reset_cache()

    def add_refver_entity(self, refver_entity, reset_cache=True, set_refver_entity_cache=True):
        # sanity check
        assert refver_entity['standard'] not in self.list_standards(), (
            f'"standard" of input RefverEntity already exists in the database.'
        )
        assert not set(refver_entity['aliases']).intersection(self.list_all_aliases()), (
            f'"aliases" of input RefverEntity has overlap with already existing ones.'
        )
        # main
        self.refver_entities.append(refver_entity)
        #if set_refver_entity_cache:
        #    refver_entity.set_cached_properties()
        if reset_cache:
            self.reset_cache()

    def reset_cache(self):
        # accessor to RefverEntity
        self.refver_entity_accessor = dict()
        for x in self.refver_entities:
            self.refver_entity_accessor[x['standard']] = x
            for alias in x['aliases']:
                self.refver_entity_accessor[alias] = x

        # aliases
        self.standard_aliases_map = dict((x['standard'], x['aliases']) for x in self.refver_entities)

        # species-to-standards
        self.species_standards_map = dict()
        for x in self.refver_entities:
            self.species_standards_map.setdefault(x['species'], list())
            self.species_standards_map[x['species']].append(x['standard'])

    # chr1 length related methods
#    @staticmethod
#    def find_chr1_name(chrom_names):
#        chr1_candidates = [x for x in chrom_names if CHR1_NAME_PAT.fullmatch(x)]
#        if len(chr1_candidates) == 1:
#            return chr1_candidates[0]
#        else:
#            raise NoChr1NameError(
#                f'Unique chromosome name which looks like chr1 could not be found.'
#            )
#
#    def find_chr1_length(self, fasta):
#        chr1_name = self.find_chr1_name(fasta.references)
#        chr1_len = fasta.lengths[fasta.references.index(chr1_name)]
#        return chr1_len

    def get_refver_entity(self, refver):
        try:
            return self.refver_entity_accessor[refver]
        except KeyError:
            raise UnknownRefverError(f'Unknown reference version: {refver}')

    # modifiers
    def change_fasta_path(self, refver, fasta_path):
        refver_entity = self.get_refver_entity(refver)
        refver_entity['fasta_path'] = fasta_path
        self.reset_cache()

    # convenience methods
    def standardize(self, refver):
        return self.get_refver_entity(refver)['standard']

    def list_standards(self):
        return list(x['standard'] for x in self.refver_entities)

    def list_all_aliases(self):
        return list(
            itertools.chain.from_iterable(
                x['aliases'] for x in self.refver_entities
            )
        )

    def list_known_refvers(self):
        return self.list_standards() + self.list_all_aliases()

    #def list_known_species(self):
    #    return tuple(set(x['species'] for x in self.refver_entities))

    def get_RefSeq_refver(self, refver):
        refver_entity = self.get_refver_entity(refver)
        RefSeq_refver = refver_entity['RefSeq_refver']
        if RefSeq_refver is None:
            raise Exception(f'This reference version does not have a matched standard RefSeq version name.')
        return RefSeq_refver

    def get_RefSeq_standard(self, refver):
        refver_entity = self.get_refver_entity(refver)
        RefSeq_refver = refver_entity['RefSeq_refver']
        species = refver_entity['species']
        if RefSeq_refver is None:
            raise Exception(f'This reference version does not have a matched standard RefSeq version name.')
        return RefSeq_refver, species
        #return {
        #    'RefSeq_refver': refver_entity['RefSeq_refver'], 
        #    'species': refver_entity['species'],
        #}

    def get_aliases(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return [refver_entity['standard']] + refver_entity['aliases']

    def find_species(self, refver):
        standard = self.standardize(refver)
        for key, val in self.species_standards_map.items():
            if standard in val:
                return key
        raise Exception(
            f'A standard refver "{standard}" could not be found from a search over "species_standards_map"'
        )

    # datafile path fetcher
#    def get_datafile_path(self, refver, map_name, refseq_pathfunc_name):
#        standard = self.standardize(refver)
#        config_path = getattr(self, map_name)[standard]
#        if config_path is None:
#            refver_entity = self.get_refver_entity(standard)
#            assert refver_entity['RefSeq_refver'] is not None
#            species = refver_entity['species']
#            result = getattr(libncbi, refseq_pathfunc_name)(standard, species)
#        else:
#            result = config_path
#        return result

    ###########################
    # reference data fetchers #
    ###########################

    def get_fasta_path(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity.processed_fasta_path

    def get_unedited_fasta_path(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity.get_unedited_fasta_path()

    def get_fasta(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity.fasta

    def get_chromdict(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity.chromdict

    def get_assemblyspec(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity.assemblyspec

    def get_chrom_converter(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity.chrom_converter

    def get_repeatmasker_path(self, refver):
    #    return self.get_datafile_path(
    #        refver, 'standard_repeatmasker_map', 'get_repeatmasker_path',
    #    )
        pass

    def get_repeat_path(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity['edited_repeat_path']

    def get_geneset_path(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity['edited_geneset_path']

    def get_regulatory_path(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity['edited_regulatory_path']

    def get_chromlenhash(self, refver):
        refver_entity = self.get_refver_entity(refver)
        return refver_entity.chromlenhash

#        standard = self.standardize(refver)
#
#        if standard not in self.cache['chromlenhash'].keys():
#            assemblyspec = self.get_assemblyspec(standard)
#            if assemblyspec is not None:
#                self.cache['chromlenhash'][standard] = ChromLengthHash.from_assemblyspec(assemblyspec)
#            else:
#                fasta_path = self.get_fasta_path(standard)
#                if fasta_path is None:
#                    self.cache['chromlenhash'][standard] = None
#                else:
#                    self.cache['chromlenhash'][standard] = ChromLengthHash.from_fasta_path(fasta_path)
#
#        return self.cache['chromlenhash'][standard]

    def get_geneset_handle(self, refver):
        standard = self.standardize(refver)
        if standard not in self.cache['geneset_handle'].keys():
            self.cache['geneset_handle'][standard] = pysam.TabixFile(
                self.get_geneset_path(standard),
                parser=pysam.asGTF(),
            )
        return self.cache['geneset_handle'][standard]

    def get_regulatory_handle(self, refver):
        standard = self.standardize(refver)
        if standard not in self.cache['regulatory_handle'].keys():
            self.cache['regulatory_handle'][standard] = pysam.TabixFile(
                self.get_regulatory_path(standard),
                parser=pysam.asGTF(),
            )
        return self.cache['regulatory_handle'][standard]

    def get_repeat_handle(self, refver):
        standard = self.standardize(refver)
        if standard not in self.cache['repeat_handle'].keys():
            self.cache['repeat_handle'][standard] = pysam.TabixFile(
                self.get_repeat_path(standard),
                parser=pysam.asBed(),
            )
        return self.cache['repeat_handle'][standard]

    # chromosome name converter
#    def get_chrom_converter(self, refver):
#        """For refver entity without a custom fasta path, 
#        ucsc-style chromosome names of assemblyspec are used.
#        """
#        # fetch refver_entity
#        refver_entity = self.access_refver_entity(refver)
#        if refver_entity['RefSeq_refver'] is None:
#            raise Exception(f'Reference version must be based on a RefSeq refver to get a chrom converter.')
#
#        # get assemblyspec
#        assemblyspec = self.get_assemblyspec(refver)
#
#        # get contigs, lengths
#        if refver_entity['fasta_path'] is None:
#            return assemblyspec.get_chrom_converter_from_self()
#        else:
#            with pysam.FastaFile(refver_entity['fasta_path']) as fasta:
#                contigs = fasta.references
#                lengths = fasta.lengths
#            return assemblyspec.get_chrom_converter(contigs, lengths)


##############
# RefverDict #
##############

class RefverDict(collections.UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        standard_list = REFVERINFO.list_standards()
        if not set(self.keys()).issubset(standard_list):
            raise Exception(
                f'RefverDict construction keys must be restricted to: '
                f'{standard_list}'
            )

    def __getitem__(self, key):
        key = REFVERINFO.standardize(key)
        try:
            result = super().__getitem__(key)
        except KeyError:
            raise Exception(f'Input reference version is not available.')

        return result

    def __contains__(self, key):
        key = REFVERINFO.standardize(key)
        return super().__contains__(key)

    def get_valid_keys(self):
        result = list()
        for standard in self.keys():
            result.extend(REFVERINFO.get_aliases(standard))
            #result.append(standard)
            #result.extend(REFVERINFO.standard_aliases_map[standard])
        return result

    ############################
    # candidates be deprecated #
    @classmethod
    def standardize(cls, refver):
        return REFVERINFO.standardize(refver)

    @classmethod
    def get_aliases(cls, refver):
        return REFVERINFO.get_aliases(refver)
    ############################


#############
# chromdict #
#############

class ChromDict(collections.OrderedDict):
    ################
    # initializers #
    ################

    def __init__(self, refver, *args, **kwargs):
        self.refver = refver
        super().__init__(*args, **kwargs)

    @classmethod
    def _init_from_pysamobj(cls, pysamobj, refver=None):
        if refver is None:
            refver = infer_refver_pysamwrapper(pysamobj)
        result = cls(refver)
        for chrom, length in zip(pysamobj.references, pysamobj.lengths):
            result[chrom] = length
        return result

    @classmethod
    def from_fasta_path(cls, fasta_path, refver=None):
        with pysam.FastaFile(fasta_path) as pysamobj:
            return cls._init_from_pysamobj(pysamobj, refver=refver)

    @classmethod
    def from_fasta(cls, fasta, refver=None):
        return cls._init_from_pysamobj(fasta, refver=refver)

    @classmethod
    def from_bam_path(cls, bam_path, refver=None):
        with pysam.AlignmentFile(bam_path) as pysamobj:
            return cls._init_from_pysamobj(pysamobj, refver=refver)

    @classmethod
    def from_bam(cls, bam, refver=None):
        return cls._init_from_pysamobj(bam, refver=refver)

    @classmethod
    def from_bamheader(cls, bamheader, refver=None):
        return cls._init_from_pysamobj(bamheader, refver=refver)

    @classmethod
    def from_vcfheader(cls, vcfheader, refver=None):
        if refver is None:
            refver = infer_refver_vcfheader(vcfheader)
        result = cls(refver)
        for contig in vcfheader.contigs.values():
            result[contig.name] = contig.length
        return result

    @classmethod
    def from_refver(cls, refver):
        fasta_path = REFVERINFO.get_fasta_path(refver)
        with pysam.FastaFile(fasta_path) as fasta:
            return cls.from_fasta(fasta, refver=refver)

    @classmethod
    def from_custom(cls, custom, refver=None):
        """Args:
            custom: {
                'contigs': ['contig1', 'contig2', ...], 
                'lengths': [length1, length2, ...],
            }
        """
        if refver is None:
            refver = infer_refver_base(custom['contigs'], custom['lengths'])
        result = cls(refver)
        for chrom, length in zip(custom['contigs'], custom['lengths']):
            result[chrom] = length
        return result

    ##############
    # properties #
    ##############

    @property
    def contigs(self):
        return list(self.keys())

    @property
    def lengths(self):
        return list(self.values())

    ##########
    # others #
    ##########

    def get_cumpos0(self, chrom, pos0):
        chromidx = self.contigs.index(chrom)
        if chromidx == 0:
            return pos0
        else:
            return pos0 + sum(self.lengths[:chromidx])

    @functools.cached_property
    def is_chr_prefixed(self):
        relevant_chroms = choose_assembled_chroms(self.contigs)
        startswith_chr = [x.startswith('chr') for x in relevant_chroms]
        if all(startswith_chr):
            return True
        elif not any(startswith_chr):
            return False
        else:
            raise Exception(f'Chromosome names are inconsistent of whether prefixed with "chr"')

    #@functools.cached_property
    def get_XY_names(self):
        def helper(name, xy):
            assert xy in ('X', 'Y')
            if name in self.contigs:
                return name
            else:
                raise NoXYError(f'No {xy} chromosome name detected')
            
        if self.is_chr_prefixed:
            X = helper('chrX', 'X')
            Y = helper('chrY', 'Y')
        else:
            X = helper('X', 'X')
            Y = helper('Y', 'Y')

        return X, Y

    @property
    def XY_names(self):
        return self.get_XY_names()

    def check_has_XY(self):
        try:
            X, Y = self.get_XY_names()
        except NoXYError:
            return False
        else:
            return True

    @functools.cached_property
    def assembled_chroms(self):
        return choose_assembled_chroms(self.contigs)

#    def to_gdf(self, assembled_only=True):
#        df = pd.DataFrame({
#            'Chromosome': self.contigs,
#            'Start': 0,
#            'End': self.lengths,
#        })
#        if assembled_only:
#            selector = df['Chromosome'].apply(
#                lambda x: PAT_ASSEMBLED_CHROM.fullmatch(x) is not None
#            )
#            df = df.loc[selector, :]
#
#        return GDF.from_frame(df, refver)

    #def to_gr(self, assembled_only=True, as_gr=True):
    #    gdf = self.to_gdf(assembled_only=assembled_only)
    #    if as_gr:
    #        return gdf.gr
    #    else:
    #        return gdf.df

#    def to_interval_list(self):
#        intvlist = libinterval.IntervalList()
#        for contig, length in zip(self.contigs, self.lengths):
#            interval = libinterval.Interval(contig, start0=0, end0=length)
#            intvlist.append(interval)
#
#        return intvlist

    def get_chrom_indexes(self, chroms):
        if np.isscalar(chroms):
            return self.contigs.index(chroms)
        else:
            where = np.where(
                np.array(chroms)[:, np.newaxis] == np.array(self.contigs)
            )
            assert len(set(where[0])) == len(chroms), f'Unknown chromosome names are included.'
            result = [
                tuple(subiter)[0][1] for key, subiter in itertools.groupby(
                    zip(where[0], where[1]), key=operator.itemgetter(0)
                )
            ]
            return result

    def get_chrompos_sortkey(self, chroms, start0s=None, end0s=None):
        assert not ((start0s is None) and (end0s is not None))

        chrom_indexes = self.get_chrom_indexes(chroms)
        if start0s is None:
            sortkey = np.lexsort([chrom_indexes])
        else:
            if end0s is None:
                sortkey = np.lexsort([start0s, chrom_indexes])
            else:
                sortkey = np.lexsort([end0s, start0s, chrom_indexes])

        return sortkey

    def sort_chrompos(self, chroms, start0s=None, end0s=None):
        assert not ((start0s is None) and (end0s is not None))

        # arg handling
        chroms = np.array(chroms)
        if start0s is not None:
            start0s = np.asarray(start0s)
        if end0s is not None:
            end0s = np.asarray(end0s)

        # result
        sortkey = self.get_chrompos_sortkey(chroms, start0s, end0s)
        if start0s is None:
            return chroms[sortkey]
        else:
            if end0s is None:
                return chroms[sortkey], start0s[sortkey]
            else:
                return chroms[sortkey], start0s[sortkey], end0s[sortkey]


############################
# Assemblyspec and related #
############################

class AssemblySpec:
    """Attributes:
        data: { 
            'length': [249250621, 243199373, ...],
            'role': ['assembled-molecule', ... 'unlocalized-scaffold', ... ],
            'default': ['1', '2', ... 'HSCHR1_RANDOM_CTG5', ... ],
            'ucsc': ['chr1', 'chr2', ... 'chr1_gl000191_random', ... ],
            #'nochr_plus_genbank': ['1', '2', ... 'GL000191.1', ... ],
            'genbank': ['CM000663.1', 'CM000664.1', ... ],
            'refseq': ['NC_000001.11', 'NC_000002.12', ... ],
        }
        dicts: {('ucsc', 'genbank'): dict, ... }
        chromdicts: {'ucsc': ChromDict object, ... }
        versions: All self.data keys except 'length' and 'role'
    """
    assembly_report_keydict = {
        'default': 'Sequence-Name',
        'ucsc': 'UCSC-style-name',
        'genbank': 'GenBank-Accn',
        'refseq': 'RefSeq-Accn',
        'length': 'Sequence-Length',
        'role': 'Sequence-Role',
    }
    def __init__(self, data):
        """
        Args:
            data: {
                'ucsc': ['chr1', 'chr2', ... ], 
                'refseq': ['NC_000001.11', 'NC_000002.12', ... ],
                ...
                }
        """

        self.data = data
        self.dicts = dict()
        #self.chromdicts = dict()

        for key1, key2 in itertools.permutations(data.keys(), 2):
            dic = dict()
            chromdict_data = {'contigs': list(), 'lengths': list()}
            for val1, val2 in zip(self.data[key1], self.data[key2]):
                if val1 is not None:
                    dic[val1] = val2
                    if key2 == 'length':
                        chromdict_data['contigs'].append(val1)
                        chromdict_data['lengths'].append(val2)

            self.dicts[(key1, key2)] = dic
            #if key2 == 'length':
            #    self.chromdicts[key1] = ChromDict(custom=chromdict_data)

        self.versions = list(
            set(self.data.keys()).difference(['length', 'role'])
        )

    @classmethod
    def from_RefSeq_refver(cls, refseq_refver, species):
        assembly_report_path = libncbicache.get_assembly_report_path(
            refseq_refver, species, force_download=False,
        )
        with open(assembly_report_path, 'r') as infile:
            result = cls.from_assembly_report_file(infile)

        return result

    @classmethod
    def from_assembly_report_file(cls, stream):
        parsed_data = cls.parse_assembly_report(stream)
        result = cls(parsed_data)
        return result

    @classmethod
    def parse_assembly_report(cls, stream):
        data = {
            'length': list(), 
            'default': list(), 
            'ucsc': list(),
            #'nochr_plus_genbank': list(), 
            'role': list(),
            'genbank': list(), 
            'refseq': list(),
        }

        record_start = False
        for line in stream:
            if not record_start:
                if line.startswith('# Sequence-Name'):
                    keys = tools.get_linesp(re.sub('^#\s*', '', line))
                    record_start = True
                    continue
            else:
                linesp = tools.get_linesp(line)
                linedict = dict(
                    zip(
                        keys, 
                        [(None if x == 'na' else x) for x in linesp]
                    )
                )

                default = linedict[cls.assembly_report_keydict['default']]
                ucsc = linedict[cls.assembly_report_keydict['ucsc']]
                genbank = linedict[cls.assembly_report_keydict['genbank']]
                refseq = linedict[cls.assembly_report_keydict['refseq']]
                length = int(linedict[cls.assembly_report_keydict['length']])
                role = linedict[cls.assembly_report_keydict['role']]

                #assert genbank is not None

                data['default'].append(default)
                data['ucsc'].append(ucsc)
                data['genbank'].append(genbank)
                data['refseq'].append(refseq)
                data['length'].append(length)
                data['role'].append(role)

#                if ucsc is None:
#                    data['nochr_plus_genbank'].append(genbank)
#                elif ucsc == 'chrM':
#                    data['nochr_plus_genbank'].append('MT')
#                else:
#                    mat = PAT_ASSEMBLED_CHROM.fullmatch(ucsc)
#                    if mat is None:
#                        data['nochr_plus_genbank'].append(genbank)
#                    else:
#                        data['nochr_plus_genbank'].append(mat.group(2))

        return data

    def __repr__(self):
        tmp_keys = list()
        tmp_vals = list()
        for key, val in self.data.items():
            tmp_keys.append(key)
            tmp_vals.append(val)

        show_result = list()
        for tup in zip(*tmp_vals):
            show_result.append(str(dict(zip(tmp_keys, tup))))

        return '\n'.join(show_result)

    def convert(self, in_chrom, out_version):
        assert out_version in self.versions, (
            f'"out_version" argument must be one of {self.versions}')

        absent = True
        for version, names in self.data.items():
            if version in ('length', 'role'):
                continue
            if in_chrom in names:
                absent = False
                if version == out_version:
                    result = in_chrom
                else:
                    result = self.dicts[(version, out_version)][in_chrom]
                break

        if absent:
            raise Exception(f'Input chrom name "{in_chrom}" is not included '
                            f'in the database.')

        return result

    def get_alias_groups(self):
        data_keys = (
            'ucsc', 
            'genbank', 
            'refseq',
            'default', 
            #'nochr_plus_genbank', 
        )
        alias_groups = list()
        for idx in range(len(self.data['length'])):
            group = set()
            for key in data_keys:
                chrom = self.data[key][idx]
                if chrom is None:
                    continue

                norm_chrom = normalize_chrom(chrom, strip_chr=False)
                norm_chrom_nochr = normalize_chrom(chrom, strip_chr=True)
                group.add(chrom)
                group.add(norm_chrom)
                group.add(norm_chrom_nochr)

            alias_groups.append(group)
        return alias_groups

    def get_chrom_converter(self, contigs, lengths):
        # check ChromLenghtHash
        hash_assemblyspec = ChromLengthHash.from_assemblyspec(self)
        hash_inputdata = ChromLengthHash(contigs, lengths)
        if not hash_assemblyspec.check_length_compatibility(hash_inputdata):
            raise Exception(f'Input contig names and lenghts does not match assemblyspec.')

        # main
        alias_groups = self.get_alias_groups()
        converter = dict()
        for chrom in contigs:
            for group in alias_groups:
                if chrom in group:
                    for alias in group:
                        assert alias not in converter
                        converter[alias] = chrom
                    continue
        return converter

    def get_chrom_converter_from_self(self, key='default'):
        """key was chosen to 'default' since some reference 
        version does not have ucsc names.
        Normalize 'default' names with 'strip_chr=False'
        """

        valid_indexes = tuple((x is not None) for x in self.data[key])
        contigs = list(itertools.compress(self.data[key], valid_indexes))
        lengths = list(itertools.compress(self.data['length'], valid_indexes))

        contigs = [normalize_chrom(x, strip_chr=False) for x in contigs]

        return self.get_chrom_converter(contigs, lengths)


#ASSEMBLYFILE_URLS = RefverDict(libncbi.collect_assemblyfile_urls())

#ASSEMBLYFILE_PATHS = {
#    key: os.path.join(ASSEMBLYFILE_DIR, os.path.basename(val))
#    for (key, val) in ASSEMBLYFILE_URLS.items()
#}


#def get_assemblyspec_data(refver):
#    assemblyfile_url = ASSEMBLYFILE_URLS[refver]
#    assembly_report_path = ASSEMBLYFILE_PATHS[refver]
#    if not os.path.exists(assembly_report_path):
#        network.download(assemblyfile_url, assembly_report_path)
#
#    return parse_assembly_report(assembly_report_path)
#
#
#SPECS = RefverDict(
#    {
#        refver: AssemblySpec(get_assemblyspec_data(refver))
#        for refver in ASSEMBLYFILE_URLS.keys()
#    }
#)


class ChromLengthHash:
    def __init__(self, chrom_list, length_list):
        self.lendict = dict(zip(chrom_list, length_list))
        self.normchrom_lendict = self.make_normchrom_lendict(chrom_list, length_list)

    @staticmethod
    def make_normchrom_lendict(chrom_list, length_list):
        normchrom_lendict = dict()
        modified_chrom_groups = dict()
        for chrom, length in zip(chrom_list, length_list):
            modified_chrom = normalize_chrom(chrom, strip_chr=False)
            normchrom_lendict[modified_chrom] = length
            modified_chrom_groups.setdefault(modified_chrom, set())
            modified_chrom_groups[modified_chrom].add(
                (chrom, length)
            )

        # sanity check
        for modified_chrom, tuples in modified_chrom_groups.items():
            lengths = set(x[1] for x in tuples)
            if len(lengths) != 1:
                raise Exception(
                    f'Modified chrom names overlap but lengths are discordant:'
                    f'modified chrom: {modified_chrom}, original chrom - length pairs: {tuples}'
                )

        return normchrom_lendict

    #@classmethod
    #def from_RefSeq_refver(cls, refseq_refver):
    #    pass

    #@classmethod
    #def from_refver(cls, refver):
    #    pass

    @classmethod
    def from_fasta(cls, fasta):
        return cls(fasta.references, fasta.lengths)

    @classmethod
    def from_fasta_path(cls, fasta_path):
        with pysam.FastaFile(fasta_path) as fasta:
            return cls.from_fasta(fasta)

    @classmethod
    def from_assemblyspec(cls, assemblyspec):
        assert isinstance(assemblyspec, AssemblySpec)

        used_keys = ('default', 'ucsc', 'genbank', 'refseq')
        chrom_list = list(
            itertools.chain.from_iterable(
                assemblyspec.data[key] for key in used_keys
            )
        )
        length_list = assemblyspec.data['length'] * len(used_keys)
        # remove None chroms
        notnone_indexes = tuple((x is not None) for x in chrom_list)
        chrom_list = list(itertools.compress(chrom_list, notnone_indexes))
        length_list = list(itertools.compress(length_list, notnone_indexes))

        return cls(chrom_list, length_list)

    @staticmethod
    def compare_lendicts(self_lendict, other_lendict):
        self_chroms = set(self_lendict.keys())
        other_chroms = set(other_lendict.keys())

        self_only_chroms = self_chroms.difference(other_chroms)
        other_only_chroms = other_chroms.difference(self_chroms)
        common_chroms = self_chroms.intersection(other_chroms)
        length_matches = all(
            (self_lendict[x] == other_lendict[x])
            for x in common_chroms
        )
        return length_matches, common_chroms, self_only_chroms, other_only_chroms

    def check_length_compatibility(self, other):
        """Used for parent standard identification"""
        assert isinstance(other, self.__class__)
        length_matches, common_chroms, self_only_chroms, other_only_chroms = (
            self.compare_lendicts(self.normchrom_lendict, other.normchrom_lendict)
        )
        return length_matches

    def check_issubset(self, other):
        assert isinstance(other, self.__class__)
        length_matches, common_chroms, self_only_chroms, other_only_chroms = (
            self.compare_lendicts(self.lendict, other.lendict)
        )
        is_subset = ((len(self_only_chroms) == 0) and length_matches)
        return is_subset, other_only_chroms

    def get_hash(self, chroms=None):
        if chroms is None:
            chrom_len_tuples = iter(self.lendict.items())
        else:
            chrom_len_tuples = ((x, self.lendict[x]) for x in chroms)
        return tuple(sorted(chrom_len_tuples, key=(lambda x: x[0])))

    def compare(self, other):
        assert isinstance(other, self.__class__)
        common_chroms = set(self.lendict.keys()).intersection(other.lendict.keys())
        return self.get_hash(common_chroms) == other.get_hash(common_chroms)
                
                    


####################
# Setup REFVERINFO #
####################

REFVERINFO = RefverInfo(handygenome.OPTION['refverinfo'], calc_checksum=False)


###########################################################################
# convenience functions for utilizing and modifying the REFVERINFO object #
###########################################################################

def standardize_refver(refver):
    return REFVERINFO.standardize(refver)


standardize = standardize_refver


def deco_standardize(func):
    sig = inspect.signature(func)
    if 'refver' not in sig.parameters.keys():
        raise Exception(
            f'parameter names must include "refver".'
        )

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ba = sig.bind(*args, **kwargs)
        ba.apply_defaults()
        ba.arguments['refver'] = standardize_refver(ba.arguments['refver'])
        return func(**ba.arguments)

    return wrapper


def add_refver_entity(refver_entity):
    REFVERINFO.add_refver_entity(refver_entity, reset_cache=True)
    

def change_fasta_path(refver, fasta_path):
    REFVERINFO.change_fasta_path(refver, fasta_path)


def list_known_refvers():
    return REFVERINFO.list_known_refvers()


def get_fasta_path(refver):
    return REFVERINFO.get_fasta_path(refver)


def get_fasta(refver):
    return REFVERINFO.get_fasta(refver)


def get_chromdict(refver):
    return REFVERINFO.get_chromdict(refver)


def get_assemblyspec(refver):
    return REFVERINFO.get_assemblyspec(refver)


def get_geneset_handle(refver):
    return REFVERINFO.get_geneset_handle(refver)
    

def get_regulatory_handle(refver):
    return REFVERINFO.get_regulatory_handle(refver)


def get_repeat_handle(refver):
    return REFVERINFO.get_repeat_handle(refver)


def get_datadir(refver):
    return REFVERINFO.get_refver_entity(refver).datadir


def get_RefSeq_standard(refver):
    return REFVERINFO.get_RefSeq_standard(refver)


def get_chrom_converter(refver):
    return REFVERINFO.get_chrom_converter(refver)


###############################
# reference version inference #
###############################

@deco.get_deco_num_set_differently(('chromdict', 'vcfheader', 'bamheader'), 1)
def infer_refver(chromdict=None, vcfheader=None, bamheader=None):
    if chromdict is not None:
        return infer_refver_chromdict(chromdict)
    elif vcfheader is not None:
        return infer_refver_vcfheader(vcfheader)
    elif bamheader is not None:
        return infer_refver_bamheader(bamheader)


#def infer_RefSeq_refver(contigs, lengths):
#    query_chromlenhash = ChromLengthHash(contigs, lengths)


def infer_refver_chromdict(chromdict):
    return infer_refver_base(chromdict.contigs, chromdict.lengths)


def infer_refver_vcfheader(vcfheader):
    contigs = list()
    lengths = list()
    for contig in vcfheader.contigs.values():
        contigs.append(contig.name)
        lengths.append(contig.length)
    return infer_refver_base(contigs, lengths)


def infer_refver_vcfpath(vcfpath):
    with pysam.VariantFile(vcfpath) as vcf:
        return infer_refver_vcfheader(vcf.header)


def infer_refver_vr(vr):
    return infer_refver_vcfheader(vr.header)


def infer_refver_fasta(fasta):
    return infer_refver_pysamwrapper(fasta)


def infer_refver_fastapath(fastapath):
    with pysam.FastaFile(fastapath) as fasta:
        return infer_refver_fasta(fasta)


def infer_refver_bamheader(bamheader):
    return infer_refver_pysamwrapper(bamheader)


def infer_refver_bampath(bampath):
    with pysam.AlignmentFile(bampath) as bam:
        return infer_refver_bamheader(bam.header)


def infer_refver_base(contigs, lengths):
    query_hash = ChromLengthHash(contigs, lengths)
    compare_results = list()
    for standard in REFVERINFO.list_standards():
        target_hash = REFVERINFO.get_chromlenhash(standard)
        #if target_hash is None:
        #    continue
        is_subset, other_only_chroms = query_hash.check_issubset(target_hash)
        refverinfo = {
            'standard': standard, 
            'is_subset': is_subset, 
            'other_only_chroms': other_only_chroms,
        }
        compare_results.append(refverinfo)

    subset_refverinfos = [x for x in compare_results if x['is_subset']]
    if len(subset_refverinfos) == 0:
        return None
    else:
        if len(subset_refverinfos) == 1:
            chosen_info = subset_refverinfos[0]
        elif len(subset_refverinfos) > 1:
            minimal_descrepancy_ones = tools.multi_min(
                subset_refverinfos, 
                key=(lambda x: len(x['other_only_chroms'])),
            )
            if len(minimal_descrepancy_ones) == 1:
                chosen_info = minimal_descrepancy_ones[0]
            else:
                raise Exception(f'Cannot break tie: {subset_refverinfos}')

        return chosen_info['standard']


def infer_refver_pysamwrapper(wrapper):
    return infer_refver_base(wrapper.references, wrapper.lengths)



##################################################
# handling of pseudoautosomal region & runs of N #
##################################################


#PARS = RefverDict(
#    # 0-based half-open system
#    # ref: https://www.ncbi.nlm.nih.gov/grc/human
#    {  
#        'GRCh37': {
#            'PAR1_X': ('X', 60_000, 2_699_520),
#            'PAR2_X': ('X', 154_931_043, 155_260_560),
#            'PAR1_Y': ('Y', 10_000, 2_649_520),
#            'PAR2_Y': ('Y', 59_034_049, 59_373_566),
#        }, 
#        'GRCh38': {
#            'PAR1_X': ('X', 10_000, 2_781_479),
#            'PAR2_X': ('X', 155_701_382, 156_030_895),
#            'PAR1_Y': ('Y', 10_000, 2_781_479),
#            'PAR2_Y': ('Y', 56_887_902, 57_217_415),
#        },
#    }
#)
#
#PAR_GRS = dict()
#for key, val in PARS.items():
#    df = pd.DataFrame.from_records(
#        iter(val.values()), 
#        columns=('Chromosome', 'Start', 'End'),
#    )
#    gr = pr.PyRanges(df).sort()
#    PAR_GRS[key] = gr
#PAR_GRS = RefverDict(PAR_GRS)
#
#
#def get_par_gr(refver):
#    chromdict = ChromDict.from_refver(refver)
#    par_gr = PAR_GRS[refver].copy()  # X, Y
#    if chromdict.is_chr_prefixed:
#        par_gr.Chromosome = 'chr' + par_gr.Chromosome
#    return par_gr





import json

import handygenome
import handygenome.network as network
import handygenome.logutils as logutils
import handygenome.fastahandler as fastahandler


URL_AUTHORITY = 'ftp.ensembl.org'
URL_PATH_GENESET = '/pub/current_gff3'
URL_PATH_REGULATION = '/pub/current_regulation'
URL_PATH_GRCH37_GENESET = '/pub/grch37/current/gff3/homo_sapiens'
URL_PATH_GRCH37_REGULATION = '/pub/grch37/current/regulation/homo_sapiens'

class EnsemblPaths:

    toppaths_cache = os.path.join(
        handygenome.DATA_DIR, 
        'ensembl_toppaths.json',
    )

    def remove_toppaths_cache(self):
        os.remove(self.__class__.toppaths_cache)

    def load_toppaths_cache(self):
        with open(self.__class__.toppaths_cache, 'rt') as f:
            self.toppaths = json.load(f)

    def save_toppaths_cache(self):
        with open(self.__class__.toppaths_cache, 'wt') as f:
            json.dump(self.toppaths, f)

    ###

    def __init__(self):
        self.initialize(force_update=False)

    def initialize(self, force_update=False):
        if force_update:
            self.remove_toppaths_cache()
        self.init_toppaths()

    def init_toppaths(self):
        if os.path.exists(self.__class__.toppaths_cache):
            self.load_toppaths_cache()
        else:
            self.toppaths = dict(
                'gff3': dict(),
                'regulation': dict(),
            )

    def set_gff3_toppaths(self):

    #####






import os
import json
import hashlib

import handygenome.logutils as logutils
import handygenome.tools as tools
import handygenome.publicdb.ncbi_cache as libncbicache
import handygenome.fastahandler as fastahandler


class PostprocessorBase:
    def __init__(
        self,
        standard_refver,
        edited_path,
        custom_path,
        RefSeq_refver,
        species,
        chrom_converter,
        unedited_file_checksum_path,
        unedited_file_wascustom_path,
        calc_checksum,
    ):
        self.standard_refver = standard_refver
        self.edited_path = edited_path
        self.custom_path = custom_path
        self.RefSeq_refver = RefSeq_refver
        self.species = species
        self.chrom_converter = chrom_converter
        self.unedited_file_checksum_path = unedited_file_checksum_path
        self.unedited_file_wascustom_path = unedited_file_wascustom_path
        self.calc_checksum = calc_checksum
        self.iscustom = (self.custom_path is not None)

    def get_helpmsg(self):
        return f'(refver={self.standard_refver}, path={repr(self.get_unedited_path())})'

    def make_processed_file(self):
        logutils.log(
            f'Postprocessing raw data file {self.get_helpmsg()}', 
            level='info',
        )
        self.postprocess(
            infile_path=self.get_unedited_path(), 
            outfile_path=self.edited_path, 
            chrom_converter=(
                self.chrom_converter
                if self.custom_path is None else
                None
            ),
        )
        logutils.log(f'Finished postprocessing', level='info')

    def prepare_processed_file(self):
        if self.calc_checksum:
            unedited_file_changed = self.compare_unedited_file_checksum()
            if unedited_file_changed:
                self.make_processed_file()
            else:
                assert os.path.exists(self.edited_path), (
                    f'Edited fasta file does not exist: {self.edited_path}'
                )
        else:
            if not os.path.exists(self.edited_path):
                self.make_processed_file()

    def compare_unedited_file_checksum(self, digest='sha256'):
        # load stored checksum which was calculated from previous opening
        if os.path.exists(self.unedited_file_checksum_path):
            with open(self.unedited_file_checksum_path, 'rt') as f:
                stored_checksum = f.read()
        else:
            stored_checksum = None

        if stored_checksum is None:
            wascustom = None
        else:
            if os.path.exists(self.unedited_file_wascustom_path):
                with open(self.unedited_file_wascustom_path, 'rt') as f:
                    wascustom = json.load(f)
            else:
                wascustom = None

        # calculate newly opened fasta checksum ; this takes a few seconds
        unedited_path = self.get_unedited_path()
        logutils.log(
            f'Calculating checksum of the unedited file. It takes a few seconds. {self.get_helpmsg()}', 
            level='info',
            add_locstring=False,
        )
        with open(unedited_path, 'rb') as f:
            new_checksum = hashlib.file_digest(f, digest).hexdigest()  
                # digest() returns bytes object, while hexdigest() returns str object

        # result
        if stored_checksum is None:
            unedited_file_changed = True
        else:
            unedited_file_changed = (stored_checksum != new_checksum) or (wascustom != self.iscustom)

        with open(self.unedited_file_checksum_path, 'wt') as f:
            f.write(new_checksum)
        with open(self.unedited_file_wascustom_path, 'wt') as f:
            json.dump(self.iscustom, f)

        return unedited_file_changed


class FastaProcessor(PostprocessorBase):
    def get_unedited_path(self):
        if self.custom_path is None:
            return libncbicache.get_fasta_path(self.RefSeq_refver, self.species)
        else:
            return self.custom_path

    @staticmethod
    def postprocess(infile_path, outfile_path, chrom_converter=None):
        if os.path.exists(outfile_path):
            os.remove(outfile_path)

        if chrom_converter is None:
            if tools.check_textfile(infile_path):
                os.symlink(infile_path, outfile_path)
            elif tools.check_gzipped(infile_path):
                tools.unzip(infile_path, outfile_path, rm_src=False)
            else:
                raise Exception(f'Input file is neither plain text nor gzipped.')
        else:
            fastahandler.rename_fasta(infile_path, outfile_path, chrom_converter)
            
        fastahandler.make_index(outfile_path)


class GenesetProcessor(PostprocessorBase):
    pass


########################
# fasta postprocessing #
########################

def prepare_processed_fasta(
    standard_refver,
    edited_path,
    custom_path,
    RefSeq_refver,
    species,
    chrom_converter,
    unedited_file_checksum_path,
    unedited_file_wascustom_path,
    calc_checksum,
):
    fasta_processor = FastaProcessor(
        standard_refver=standard_refver,
        edited_path=edited_path,
        custom_path=custom_path,
        RefSeq_refver=RefSeq_refver,
        species=species,
        chrom_converter=chrom_converter,
        unedited_file_checksum_path=unedited_file_checksum_path,
        unedited_file_wascustom_path=unedited_file_wascustom_path,
        calc_checksum=calc_checksum,
    )
    fasta_processor.prepare_processed_file()

###########
# geneset #
###########

def prepare_processed_geneset(self):
    pass






import os
import warnings
import re
import logging

import pysam

import handygenome.annotation.ensembl_feature as ensembl_feature
import handygenome.variant.vcfspec as libvcfspec


DEFAULT_FETCHWIDTH = 10

PAT_GENESET_ID = re.compile('ID=(?P<type>[^:]+):(?P<id>[^;]+)')
    #   type: CDS, gene, transcript (obtained from: 
    #        zcat geneset_gff3_sorted.gz | 
    #            awk '
    #            {
    #                if ($9 ~ /ID=([^:]+):/) {
    #                    print gensub(/^.*ID=([^:]+):.*$/, "\\1", "g", $9)
    #                }
    #            }' | 
    #            sort | uniq)
PAT_GENESET_PARENT = re.compile('Parent=[^:]+:(?P<id>[^;]+)')
PAT_REGULATORY_ID = re.compile('(.+;)?id=([^;]+)(;.+)?')


# VCF

class NoVcfIndexError(Exception):
    pass


def fetch_relevant_vr_multialt(vcfspec, vcf, search_equivs=True, raise_with_multihit=True):
    result = list()
    for vcfspec_monoalt in vcfspec.iter_monoalts():
        result.append(
            fetch_relevant_vr(
                vcfspec_monoalt, 
                vcf, 
                search_equivs=search_equivs, 
                raise_with_multihit=raise_with_multihit,
            )
        )

    return result


def fetch_relevant_vr(vcfspec, vcf, search_equivs=True, raise_with_multihit=True):
    """Args:
        vcfspec: Must be with one ALT allele.
        vcf: pysam.VariantFile object
        search_equivs: If True, all equivalent forms are fetched. If False, 
            only vrs with identical chrom, pos, ref, alt are fetched.
    """

    def matcher_exact(query_vcfspec, target_vcfspec):
        return query_vcfspec == target_vcfspec

    def matcher_any(query_vcfspec, target_vcfspec):
        return any((query_vcfspec == x) for x in target_vcfspec.iter_monoalts())

    def search_equivalents(vcfspec, vcf, matcher):
        if vcfspec.chrom in vcf.header.contigs:
            equivs = vcfspec.get_equivalents()
            poslist = [x.pos for x in equivs]
            fetchresult = vcf.fetch(contig=vcfspec.chrom, 
                                    start=(min(poslist) - 1),
                                    end=max(poslist))

            relevant_vr_candidates = list()
            for vr in fetchresult:
                vr_vcfspec = libvcfspec.make_parsimonious(
                    libvcfspec.Vcfspec.from_vr(vr)
                )
                if any(matcher(query, vr_vcfspec) for query in equivs):
                    relevant_vr_candidates.append(vr)
        else:
            relevant_vr_candidates = list()

        return relevant_vr_candidates

    def search_identical(vcfspec, vcf, matcher):
        if vcfspec.chrom in vcf.header.contigs:
            fetchresult = vcf.fetch(contig=vcfspec.chrom,
                                    start=(vcfspec.pos - 1), end=vcfspec.pos)

            relevant_vr_candidates = list()
            for vr in fetchresult:
                vr_vcfspec = libvcfspec.Vcfspec.from_vr(vr)
                if matcher(vcfspec, vr_vcfspec):
                    relevant_vr_candidates.append(vr)
        else:
            relevant_vr_candidates = list()
    
        return relevant_vr_candidates

    def get_relevant_vr(relevant_vr_candidates, raise_with_multihit):
        if len(relevant_vr_candidates) == 0:
            relevant_vr = None
        elif len(relevant_vr_candidates) == 1:
            relevant_vr = relevant_vr_candidates[0]
        else:
            if raise_with_multihit:
                e_msg = list()
                e_msg.append(f'There are more than one relevant variant records:')
                for vr in relevant_vr_candidates:
                    e_msg.append(str(vr))
                raise Exception('\n'.join(e_msg))
            else:
                relevant_vr = relevant_vr_candidates[0]

        return relevant_vr

    # main
    vcfspec.check_monoalt(raise_with_false=True)

    if vcf.index is None:
        raise NoVcfIndexError(f'Input vcf is not indexed.')

    matcher = matcher_exact

    if search_equivs:
        relevant_vr_candidates = search_equivalents(vcfspec, vcf, matcher)
    else:
        relevant_vr_candidates = search_identical(vcfspec, vcf, matcher)

    relevant_vr = get_relevant_vr(relevant_vr_candidates, raise_with_multihit)

    return relevant_vr


def fetch_relevant_vr_vcfpath(vcfspec, vcf_path, search_equivs=True):
    with pysam.VariantFile(vcf_path, 'r') as vcf:
        try:
            relevant_vr = fetch_relevant_vr(vcfspec, vcf, search_equivs=search_equivs)
        except NoVcfIndexError:
            print(f'Input vcf path: {vcf_path}')
            raise

    return relevant_vr


# cosmic


# geneset gff3

def parse_transcript_tabixline(tabixline):
    transcript = ensembl_feature.Transcript(is_missing=False)

    attrs_parsed = parse_gff3_attrs(tabixline.attributes)
    transcript['id'] = attrs_parsed['ID'].split(':')[1]
    transcript['biotype'] = attrs_parsed['biotype']

    if tabixline.strand == '+':
        transcript['is_forward'] = True
    elif tabixline.strand == '-':
        transcript['is_forward'] = False
    else:
        raise Exception(f'Unexpected transcript tabixline strand value: '
                        f'"{tabixline.strand}"')

    transcript['chrom'] = tabixline.contig
    transcript['start0'] = tabixline.start
    transcript['start1'] = transcript['start0'] + 1
    transcript['end0'] = tabixline.end
    transcript['end1'] = transcript['end0']

    if 'Name' in attrs_parsed:
        transcript['transcript_name'] = attrs_parsed['Name']
    else:
        transcript['transcript_name'] = None

    transcript['gene_id'] = attrs_parsed['Parent'].split(':')[1]

    return transcript


def fetch_transcript(chrom, start0, end0, tabixfile_geneset):
    """Returns an empty TranscriptSet object if the tabixfile does not 
    contain chrom.
    """
    transcript_set = ensembl_feature.TranscriptSet(is_missing=False)
    if chrom in tabixfile_geneset.contigs:
        try:
            fetchresult = tabixfile_geneset.fetch(chrom, start0, end0)
        except Exception as e:
            raise Exception(f'tabixfile fetch failed: chrom: {chrom}, '
                            f'start0: {start0}, end0: {end0}') from e
            
        for tabixline in fetchresult:
            if check_tabixline_is_transcript(tabixline):
                transcript = parse_transcript_tabixline(tabixline)
                transcript_set[transcript['id']] = transcript

    return transcript_set


def fetch_transcript_tabixline(chrom, start0, end0, transcript_id_list, 
                               tabixfile_geneset):
    if chrom not in tabixfile_geneset.contigs:
        raise Exception(f'chrom is not contained in the tabixfile.')

    candidates = list()
    fetched = tabixfile_geneset.fetch(chrom, start0, end0)
    for tabixline in fetched:
        mat = PAT_GENESET_ID.search(tabixline.attributes)
        if mat is not None:
            if mat.group('type') == 'transcript':
                if mat.group('id') in transcript_id_list:
                    candidates.append(tabixline)

    if len(candidates) != len(transcript_id_list):
        raise Exception(f'Unsuccessful transcript fetch: '
                        f'transcript_id_list: {transcript_id_list}, '
                        f'coords: ({chrom}, {start0}, {end0})')

    return candidates


def check_geneset_tabixline_type(tabixline, type):
    assert type in ('gene', 'transcript', 'CDS', 'exon', 'UTR'), (
        f'Allowed "type" values are: '
        f'"gene", "transcript", "CDS", "exon", "UTR"')

    if type == 'exon':
        return tabixline.feature == 'exon'
    elif type == 'UTR':
        return tabixline.feature in ('five_prime_UTR', 'three_prime_UTR')
    else:
        mat = PAT_GENESET_ID.search(tabixline.attributes)
        if mat is None:
            return False
        else:
            return (mat.group('type') == type)


def check_tabixline_is_transcript(tabixline):
    mat = PAT_GENESET_ID.search(tabixline.attributes)
    if mat is None:
        return False
    else:
        return (mat.group('type') == 'transcript')


def get_parent(tabixline):
    mat = PAT_GENESET_PARENT.search(tabixline.attributes)
    if mat is None:
        return None
    else:
        return mat.group('id')


# repeats bed

def parse_repeat_tabixline(tabixline):
    repeat = ensembl_feature.Repeat(is_missing=False)

    repeat['chrom'] = tabixline.contig
    repeat['start0'] = tabixline.start
    repeat['end0'] = tabixline.end
    repeat['start1'] = repeat['start0'] + 1
    repeat['end1'] = repeat['end0']

    raw_split = tabixline.name.split(':')

    clsfml = raw_split[0].split('/')
    if len(clsfml) == 1:
        repeat['class'] = clsfml[0]
        repeat['family'] = clsfml[0]
    elif len(clsfml) == 2:
        repeat['class'] = clsfml[0]
        repeat['family'] = clsfml[1]
    else:
        raise Exception(f'Unexpected repeat name pattern: {tabixline.name}')

    repeat['name'] = raw_split[1]

    return repeat


def fetch_repeat(chrom, start0, end0, tabixfile_repeats):
    """Returns an empty EnsemblFeatureSet object if the tabixfile does not 
    contain chrom.
    """
    repeat_set = ensembl_feature.RepeatSet(is_missing=False)
    if chrom in tabixfile_repeats.contigs:
        try:
            fetchresult = tabixfile_repeats.fetch(chrom, start0, end0)
        except Exception as e:
            raise Exception(f'tabixfile fetch failed: chrom: {chrom}, '
                            f'start0: {start0}, end0: {end0}') from e

        for tabixline in fetchresult:
            repeat = parse_repeat_tabixline(tabixline)
            repeat_set[repeat['id']] = repeat

    return repeat_set


# regulatory gff3

def parse_regulatory_tabixline(tabixline):
    regulatory = ensembl_feature.Regulatory(is_missing=False)

    regulatory['chrom'] = tabixline.contig
    regulatory['start0'] = tabixline.start
    regulatory['end0'] = tabixline.end
    regulatory['start1'] = regulatory['start0'] + 1
    regulatory['end1'] = regulatory['end0']
    regulatory['biotype'] = tabixline.feature

    attrs_parsed = parse_gff3_attrs(tabixline.attributes)
    regulatory['id'] = attrs_parsed['id']
    regulatory['activity'] = dict(
        x.split(':') for x in attrs_parsed['activity'].split(',')
    )

    regulatory['bound_start1'] = int(attrs_parsed['bound_start'])
    regulatory['bound_end1'] = int(attrs_parsed['bound_end'])
    regulatory['bound_start0'] = regulatory['bound_start1'] - 1
    regulatory['bound_end0'] = regulatory['bound_end1']

    return regulatory


def fetch_regulatory_tabixline(chrom, start0, end0, regulatory_id_list, 
                               tabixfile_regulatory):
    if chrom not in tabixfile_regulatory.contigs:
        raise Exception(f'chrom is not contained in the tabixfile.')

    candidates = list()
    fetched = tabixfile_regulatory.fetch(chrom, start0, end0)
    for tabixline in fetched:
        ID = get_ID_regulatory_tabixline(tabixline)
        if ID in regulatory_id_list:
            candidates.append(tabixline)

    if len(candidates) != len(regulatory_id_list):
        raise Exception(f'Unsuccessful regulatory fetch: '
                        f'regulatory_id_list: {regulatory_id_list}, '
                        f'coords: ({chrom}, {start0}, {end0})')

    return candidates


def fetch_regulatory(chrom, start0, end0, tabixfile_regulatory):
    """Returns an empty AnnotItemDict object if the tabixfile does not 
    contain chrom.
    """

    regulatory_set = ensembl_feature.RegulatorySet(is_missing=False)
    if chrom in tabixfile_regulatory.contigs:
        try:
            fetchresult = tabixfile_regulatory.fetch(chrom, start0, end0)
        except Exception as e:
            raise Exception(f'tabixfile fetch failed: chrom: {chrom}, '
                            f'start0: {start0}, end0: {end0}') from e

        for tabixline in fetchresult:
            regulatory = parse_regulatory_tabixline(tabixline)
            regulatory_set.add_feature(regulatory)

    return regulatory_set


def get_ID_regulatory_tabixline(tabixline):
    return PAT_REGULATORY_ID.search(tabixline.attributes).group(2)

# misc.

def parse_gff3_attrs(raw_string):
    return dict(x.split('=') for x in raw_string.split(';'))




import os
import tempfile
import subprocess
import shutil

import pyranges as pr
import numpy as np
import pandas as pd

import handygenome
import handygenome.workflow as workflow
import handygenome.ucscdata as ucscdata
import handygenome.pyranges_helper as pyranges_helper
import handygenome.cnv.misc as cnvmisc


LOGGER_INFO = workflow.get_debugging_logger(verbose=False)
LOGGER_DEBUG = workflow.get_debugging_logger(verbose=True)

RSCRIPT_COPYNUMBER = '/home/users/pjh/tools/miniconda/221104/miniconda3/envs/r-copynumber/lib/R/bin/Rscript'
COPYNUMBER_SCRIPT_PATH = os.path.join(handygenome.DIRS['R'], 'copynumber.R')


def add_arm_info(gr, cytoband_gr, as_gr=True):
    if isinstance(gr, pd.DataFrame):
        gr = pr.PyRanges(gr, int64=False)

    new_cytoband_gr = pr.from_dict(
        {
            'Chromosome': cytoband_gr.Chromosome,
            'Start': cytoband_gr.Start,
            'End': cytoband_gr.End,
            'arm': [x[0] for x in cytoband_gr.Name],
        },
        int64=False,
    )

    joined_gr = pyranges_helper.join(gr, new_cytoband_gr, how='left', merge='first')
    if as_gr:
        return joined_gr
    else:
        return joined_gr.df


def add_arm_info_new(df, cytoband_df, refver, as_gr=False):
    new_cytoband_df = cytoband_df.loc[:, ['Chromosome', 'Start', 'End']]
    new_cytoband_df['arm'] = [x[0] for x in cytoband_df.Name]
    joined_df = pyranges_helper.join(
        df, new_cytoband_df, how='left', merge='first',
        sort=True, refver=refver,
    )

    if as_gr:
        return pr.PyRanges(joined_df)
    else:
        return joined_df


def merge_depth_baf_respect_bafs(depth_df, baf_df):
    depth_gr = pr.PyRanges(depth_df, int64=False)
    baf_gr = pr.PyRanges(baf_df, int64=False)

    result = pyranges_helper.isec_union(depth_gr, baf_gr)
    result = pyranges_helper.join(result, depth_gr, how='left', merge='first')
        # Now "depth" column is added
    result = pyranges_helper.join(result, baf_gr, how='left', merge='first')
        # Now "baf" column is added. Missing values are np.nan

    return result


def merge_depth_baf_average_bafs(depth_df, baf_df):
    depth_gr = pr.PyRanges(depth_df, int64=False)
    baf_gr = pr.PyRanges(baf_df, int64=False)
    result = pyranges_helper.join_new(depth_gr, baf_gr, how='left', merge='mean')

    return result


def make_merged_gr(depth_df, baf_df, cytoband_gr=None, refver=None):
    if cytoband_gr is None:
        cytoband_gr = ucscdata.get_cytoband_gr(refver)

    if baf_df is None:
        merged_gr = pr.PyRanges(depth_df, int64=False)
    else:
        merged_gr = merge_depth_baf_average_bafs(depth_df, baf_df)
        if getattr(merged_gr, 'baf_raw').isna().all():
            merged_gr = merged_gr[merged_gr.columns.drop(['baf_raw']).to_list()]

    merged_gr = add_arm_info(merged_gr, cytoband_gr, as_gr=True)

    selector = np.logical_and(
        getattr(merged_gr, 'depth_raw').notna().to_numpy(), 
        getattr(merged_gr, 'arm').notna().to_numpy(),
    )
    merged_gr_wona = merged_gr[pd.Series(selector)].sort()

    return merged_gr, merged_gr_wona


def make_merged_df(depth_df, baf_df, cytoband_df, refver, logger, nproc):
    if cytoband_df is None:
        cytoband_df = ucscdata.get_cytoband_gr(refver=refver, as_gr=False)

    if baf_df is None:
        merged_df = depth_df
    else:
        logger.debug(f'Joining depth and baf dataframes')
        merged_df = pyranges_helper.join(
            depth_df, baf_df, 
            how='left', 
            merge='mean', 
            find_nearest=False,
            as_gr=False,
            sort=True,
            refver=refver,
            verbose=True,
            nproc=nproc,
        )
        if merged_df['baf_raw'].isna().all():
            merged_df = merged_df.drop('baf_raw')

    logger.debug(f'Adding chromosome arm information')
    merged_df = add_arm_info_new(merged_df, cytoband_df, refver, as_gr=False)

    logger.debug(f'Removing rows with missing depth or arm information')
    selector = np.logical_and(
        merged_df['depth_raw'].notna().to_numpy(), 
        merged_df['arm'].notna().to_numpy(), 
    )
    merged_df_wona = merged_df.loc[selector, :]

    return merged_df, merged_df_wona


#def make_compact(merged_gr):
#    poslist = list(range(1, merged_gr.df.shape[0] + 1))  # starts with 1
#    key_iter = zip(merged_gr.Start, merged_gr.End)
#    coord_converter = dict(zip(key_iter, iter(poslist)))
#    rev_coord_converter = {val: key for key, val in coord_converter.items()}
#
#    compacted_df = pd.DataFrame.from_dict({
#        'chrom': merged_gr.Chromosome.array,
#        'pos': poslist,
#        'arm': merged_gr.arm.array,
#        'depth': getattr(merged_gr, 'depth_raw').array,
#    })
#    if 'baf_raw' in merged_gr.columns:
#        compacted_df['baf'] = getattr(merged_gr, 'baf_raw').array
#
#    return compacted_df, coord_converter, rev_coord_converter


def make_compact_df(merged_df, refver):
    assert isinstance(merged_df, pd.DataFrame)

    # sort input df
    merged_df = cnvmisc.sort_genome_df(merged_df, refver)

    # make mapping
    compact_pos1_list = list()

    original_to_compact = dict()
    for chrom, subdf in merged_df.groupby(merged_df['Chromosome'].to_numpy()):
        compact_pos1_sublist = np.arange(subdf.shape[0]) + 1
        key_iter = zip(subdf['Start'], subdf['End'])
        original_to_compact[chrom] = dict(zip(key_iter, compact_pos1_sublist))
        compact_pos1_list.extend(compact_pos1_sublist)

    compact_to_original = dict()
    for chrom, subdic in original_to_compact.items():
        compact_to_original[chrom] = {
            compact_pos1: original_tup for 
            original_tup, compact_pos1 in subdic.items()
        }

    # make compact df
    compact_df = pd.DataFrame.from_dict(
        {
            'chrom': merged_df.Chromosome.array,
            'pos': compact_pos1_list,
            'arm': merged_df.arm.array,
            'depth': merged_df.depth_raw.array,
        }
    )
    if 'baf_raw' in merged_df.columns:
        compact_df['baf'] = merged_df['baf_raw'].array

    return compact_df, compact_to_original


def make_noncompact_input(merged_gr):
    df_written = pd.DataFrame.from_dict({
        'chrom': merged_gr.Chromosome.array,
        'pos': (((merged_gr.Start + 1) + merged_gr.End) / 2).astype('int').array,
            # to make into 1-based coordinate
        'arm': merged_gr.arm.array,
        'depth': getattr(merged_gr, 'depth_raw').array,
    })
    if 'baf_raw' in merged_gr.columns:
        df_written['baf'] = getattr(merged_gr, 'baf_raw').array

    return df_written


def make_noncompact_input_df(merged_df):
    df_written = pd.DataFrame.from_dict({
        'chrom': merged_df.Chromosome.array,
        'pos': (((merged_df.Start + 1) + merged_df.End) / 2).astype('int').array,
            # to make into 1-based coordinate
        'arm': merged_df.arm.array,
        'depth': merged_df.depth_raw.array,
    })
    if 'baf_raw' in merged_df.columns:
        df_written['baf'] = merged_df.baf_raw.array

    return df_written


def load_compact_rcopynumber_output(outfile_path, compact_to_original):
    # load text file
    raw_seg = pd.read_table(
        outfile_path, 
        sep='\t', 
        header=0, 
        dtype={'chrom': str},
    )

    # trim segment limits
    for chrom in set(raw_seg['chrom']):
        compact_pos1_min = min(compact_to_original[chrom].keys())
        compact_pos1_max = max(compact_to_original[chrom].keys())
        selector = (raw_seg['chrom'] == chrom)
        raw_seg.loc[selector, 'start.pos'] = np.clip(
            raw_seg.loc[selector, 'start.pos'], 
            compact_pos1_min, 
            compact_pos1_max,
        )
        raw_seg.loc[selector, 'end.pos'] = np.clip(
            raw_seg.loc[selector, 'end.pos'], 
            compact_pos1_min, 
            compact_pos1_max,
        )

    raw_seg.drop_duplicates(
        ['chrom', 'start.pos', 'end.pos'], 
        keep='first', 
        inplace=True,
    )

    # set parameters
    if 'logR.mean' in raw_seg.columns:
        depthlist = raw_seg['logR.mean'].array
        baflist = raw_seg['BAF.mean'].array
    else:
        depthlist = raw_seg['mean'].array
        baflist = None
                
    # do conversion
    def start_converter(row):
        return compact_to_original[row['chrom']][row['start.pos']][0]

    def end_converter(row):
        return compact_to_original[row['chrom']][row['end.pos']][1]

    start0_list = raw_seg.loc[:, ['chrom', 'start.pos']].apply(start_converter, axis=1)
    end0_list = raw_seg.loc[:, ['chrom', 'end.pos']].apply(end_converter, axis=1)

    assert start0_list.notna().all()
    assert end0_list.notna().all()

    # make result
    source_dict = {
        'Chromosome': raw_seg['chrom'].array, 
        'Start': start0_list.array,
        'End': end0_list.array,
        'depth_segment_mean': depthlist,
    }
    if baflist is not None:
        source_dict['baf_segment_mean'] = baflist

    seg = pd.DataFrame.from_dict(source_dict)

    return seg


def sanitycheck(
    argtype1, 
    argtype2, 
    argtype3,
    required_columns,
    depth_df,
    baf_df,
    refver, 
    cytoband_gr, 
    merged_df, 
    merged_df_wona,
):
    assert refver is not None, f'"refver" must be given'

    if depth_df is not None:
        if not set(required_columns['depth']).issubset(depth_df.columns):
            raise Exception(
                f'"depth_df" columns must include all of these: {required_columns["depth"]}'
            )
    if baf_df is not None:
        if not set(required_columns['baf']).issubset(baf_df.columns):
            raise Exception(
                f'"baf_df" columns must include all of these: {required_columns["baf"]}'
            )

    if argtype1 or argtype3:
        if (refver is None) and (cytoband_gr is None):
            raise Exception(f'At least one of "refver" or "cytoband_gr" argument must be set.')
    elif argtype2:
        if (merged_df is None) or (merged_df_wona is None):
            raise Exception(f'Both "merged_df", and "merged_df_wona" must be given.')


def input_df_handling(input_df, remove_unassembled_contigs, req_cols):
    if input_df is None:
        return None
    else:
        input_df = input_df.copy()
        input_df = cnvmisc.arg_into_df(input_df)
        if remove_unassembled_contigs:
            input_df = cnvmisc.remove_unassembled_contigs(input_df)

        # remove unnecessary columns
        input_df = input_df.loc[:, req_cols]

        return input_df


def argtype1_prepare_input(
    depth_df, baf_df, remove_unassembled_contigs, required_columns,
    cytoband_gr, refver, logger, join_nproc,
):
    depth_df = input_df_handling(
        depth_df, remove_unassembled_contigs, required_columns['depth'],
    )
    baf_df = input_df_handling(
        baf_df, remove_unassembled_contigs, required_columns['baf'],
    )

    # make input data for r-copynumber
    merged_df, merged_df_wona = make_merged_df(
        depth_df, baf_df, cytoband_gr, refver, logger, join_nproc,
    )

    return merged_df, merged_df_wona


def argtype3_prepare_input(
    baf_df, remove_unassembled_contigs, required_columns,
    cytoband_gr, refver, logger, join_nproc,
):
    merged_df = input_df_handling(
        baf_df, remove_unassembled_contigs, required_columns['baf'],
    )
    merged_df['depth_raw'] = 1

    if cytoband_gr is None:
        cytoband_gr = ucscdata.get_cytoband_gr(refver=refver, as_gr=False)
    merged_df = add_arm_info_new(merged_df, cytoband_gr, refver, as_gr=False)

    selector = merged_df['arm'].notna().to_numpy()
    merged_df_wona = merged_df.loc[selector, :]

    return merged_df, merged_df_wona


###########################################################################


def run_rcopynumber(
    *,
    depth_df=None, 
    baf_df=None, 
    refver=None, 
    cytoband_gr=None, 

    merged_df=None, 
    merged_df_wona=None,

    as_gr=True, 
    winsorize=False, 
    compact=False,
    verbose=False, 
    remove_unassembled_contigs=True,
    join_nproc=1,
    **kwargs,

    #df_written=None,
):
    """Args:
        depth_df:
            columns: Chromosome, Start, End, depth_raw
        baf_df:
            columns: Chromosome, Start, End, baf_raw
        refver: A required argument

    Returns:
        seg: Columns ("Chromosome", "Start", "End", "depth_segment_mean", +/- "baf_segment_mean")
        merged_df: Columns ("Chromosome", "Start", "End", "depth_raw", +/- "baf_raw")
    """
    # argtype
    argtype1 = (depth_df is not None)
    argtype2 = (merged_df is not None) or (merged_df_wona is not None)
    argtype3 = (depth_df is None) and (baf_df is not None)
    if (argtype1 + argtype2 + argtype3) != 1:
        raise Exception(f'Invalid argument usage')

    # sanity checks
    required_columns = {
        'depth': ['Chromosome', 'Start', 'End', 'depth_raw'],
        'baf': ['Chromosome', 'Start', 'End', 'baf_raw'],
    }
    sanitycheck(
        argtype1, 
        argtype2, 
        argtype3,
        required_columns,
        depth_df,
        baf_df,
        refver, 
        cytoband_gr, 
        merged_df, 
        merged_df_wona,
    )

    # set logger
    logger = (LOGGER_DEBUG if verbose else LOGGER_INFO)

    # make tmpdir
    tmpdir = tempfile.mkdtemp(dir=os.getcwd(), prefix=f'tmpdir_Rcopynumber_')

    if argtype1:
        merged_df, merged_df_wona = argtype1_prepare_input(
            depth_df, baf_df, remove_unassembled_contigs, required_columns,
            cytoband_gr, refver, logger, join_nproc,
        )
    elif argtype3:
        merged_df, merged_df_wona = argtype3_prepare_input(
            baf_df, remove_unassembled_contigs, required_columns,
            cytoband_gr, refver, logger, join_nproc,
        )
    elif argtype2:
        pass

    # write input file for R copynumber
    if compact:
        logger.debug(f'Making compact')
        df_written, compact_to_original = make_compact_df(merged_df_wona, refver)
    else:
        df_written = make_noncompact_input_df(merged_df_wona)

    logger.debug(f'Writing dataframe into file for R-copynumber input')
    infile_path = os.path.join(tmpdir, 'input.tsv.gz')
    df_written.to_csv(infile_path, sep='\t', index=False)

    # run R copynumber
    logger.debug(f'Running R-copynumber')
    args = [RSCRIPT_COPYNUMBER, COPYNUMBER_SCRIPT_PATH, '--infile', infile_path]
    if compact:
        args.append('--asis')

    if winsorize:
        args.append('--winsorize')
    if verbose:
        args.append('--verbose')

    for key, val in kwargs.items():
        if val is not None:
            args.append(f'--{key}')
            args.append(str(val))

    p = subprocess.run(args, capture_output=False)  
        # "capture_output=False" lets stdout and stderr to be printed immediately
    p.check_returncode()
        
    # load copynumber output
    logger.debug(f'Loading R-copynumber output')
    outfile_path = infile_path + '.out.gz'
    if compact:
        seg = load_compact_rcopynumber_output(outfile_path, compact_to_original)
    else:
        seg = pd.read_table(
            outfile_path, sep='\t', header=0, 
            dtype={'Chromosome': str, 'Start': int, 'End': int},
        )

    # remove tmpdir
    shutil.rmtree(tmpdir)

    # return
    if as_gr:
        seg = pr.PyRanges(seg, int64=False)
        merged_df = pr.PyRanges(merged_df, int64=False)
    #else:
        #merged_gr = merged_gr.df

    return seg, merged_df

run_rcopynumber_unified = run_rcopynumber


def add_CNn_to_targetseq_segment_gr(segment_gr, targetregion_gr, refver, is_female, as_gr=True):
    """With targeted sequencing analysis, start and end coordinates of 
    "run_rcopynumber*" function output does not match actual target regions.
    To add CNn values to such a segment PyRanges, 1) CNn values are added
    to the target region PyRanges, 2) segment PyRanges is joined to the 
    annotated target region PyRanges to acquire CNn values
    Doing like this can handle an exceptional case where non-existent segment 
    subregion (a region included in the segment df but not part of the target 
    regions) spans PAR.
    """
    targetregion_gr = pyranges_helper.join(
        targetregion_gr, cnvmisc.get_CNn_gr(refver, is_female),
        how='left', merge='first', as_gr=False,
    )
    segment_gr = pyranges_helper.join(
        segment_gr, targetregion_gr, how='left', merge='longest', as_gr=as_gr,
        sort=True, refver=refver,
    )
    return segment_gr

add_CNn_to_targetseq_segment = add_CNn_to_targetseq_segment_gr


def add_CNn_to_wgs_segment_gr(segment_gr, refver, is_female, as_gr=True):
    return pyranges_helper.join(
        segment_gr, cnvmisc.get_CNn_gr(refver, is_female), 
        how='left', merge='longest', as_gr=as_gr,
        sort=True, refver=refver,
    )

add_CNn_to_wgs_segment = add_CNn_to_wgs_segment_gr





import os
import warnings
import re
import logging

import pysam

import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))
workflow = importlib.import_module('.'.join([top_package_name, 'workflow']))
infoformat = importlib.import_module('.'.join([top_package_name, 'variantplus', 'infoformat']))
equivalents = importlib.import_module('.'.join([top_package_name, 'variantplus', 'equivalents']))
varianthandler = importlib.import_module('.'.join([top_package_name, 'variantplus', 'varianthandler']))
annotationdb = importlib.import_module('.'.join([top_package_name, 'annotation', 'annotationdb']))
ensembl_parser = importlib.import_module('.'.join([top_package_name, 'annotation', 'ensembl_parser']))


DEFAULT_FETCHWIDTH = 10

PAT_GENESET_ID = re.compile('ID=(?P<type>[^:]+):(?P<id>[^;]+)')
    #   type: CDS, gene, transcript (obtained from: 
    #        zcat geneset_gff3_sorted.gz | 
    #            awk '
    #            {
    #                if ($9 ~ /ID=([^:]+):/) {
    #                    print gensub(/^.*ID=([^:]+):.*$/, "\\1", "g", $9)
    #                }
    #            }' | 
    #            sort | uniq)
PAT_GENESET_PARENT = re.compile('Parent=[^:]+:(?P<id>[^;]+)')
PAT_REGULATORY_ID = re.compile('(.+;)?id=([^;]+)(;.+)?')


# VCF

class NoVcfIndexError(Exception):
    pass


def fetch_relevant_vr_multialt(vcfspec, vcf, fasta=None, search_equivs=True,
                               allow_multiple=False):
    result = list()
    for alt in vcfspec.alts:
        new_vcfspec = common.Vcfspec(vcfspec.chrom, vcfspec.pos, vcfspec.ref, (alt,))
        result.append(
            fetch_relevant_vr(new_vcfspec, vcf, fasta=fasta, search_equivs=search_equivs,
                              allow_multiple=allow_multiple)
        )

    return result


def fetch_relevant_vr(vcfspec, vcf, fasta=None, search_equivs=True,
                      allow_multiple=False):
    """Args:
        vcfspec: Must be with one ALT allele.
        vcf: pysam.VariantFile object
        fasta: Not required when 'search_equivs' is False
        search_equivs: If True, all equivalent forms are fetched. If False, 
            only vrs with identical chrom, pos, ref, alt are fetched.
    """

    assert not ((fasta is None) and search_equivs), (
        f'If "search_equivs" is True, "fasta" must be given.')
    assert len(vcfspec.alts) == 1, (
        f'Argument "vcfspec" must be with one ALT allele.')

    def matcher_exact(query_vcfspec, target_vcfspec):
        return query_vcfspec == target_vcfspec

    def matcher_any(query_vcfspec, target_vcfspec):
        return any((query_vcfspec == x) for x in target_vcfspec.iter_monoalts())

    def search_equivalents(vcfspec, vcf, matcher, fasta):
        if vcfspec.chrom in vcf.header.contigs:
            equivs = equivalents.indel_equivalents(vcfspec, fasta)
            poslist = [x.pos for x in equivs]
            fetchresult = vcf.fetch(contig=vcfspec.chrom, 
                                    start=(min(poslist) - 1),
                                    end=max(poslist))

            relevant_vr_candidates = list()
            for vr in fetchresult:
                vr_vcfspec = equivalents.make_parsimonious(
                    varianthandler.get_vcfspec(vr)
                )
                if any(matcher(query, vr_vcfspec) for query in equivs):
                    relevant_vr_candidates.append(vr)
        else:
            relevant_vr_candidates = list()

        return relevant_vr_candidates

    def search_identical(vcfspec, vcf, matcher):
        if vcfspec.chrom in vcf.header.contigs:
            fetchresult = vcf.fetch(contig=vcfspec.chrom,
                                    start=(vcfspec.pos - 1), end=vcfspec.pos)

            relevant_vr_candidates = list()
            for vr in fetchresult:
                vr_vcfspec = varianthandler.get_vcfspec(vr)
                if matcher(vcfspec, vr_vcfspec):
                    relevant_vr_candidates.append(vr)
        else:
            relevant_vr_candidates = list()
    
        return relevant_vr_candidates

    def get_relevant_vr(relevant_vr_candidates, allow_multiple):
        if len(relevant_vr_candidates) == 0:
            relevant_vr = None
        elif len(relevant_vr_candidates) == 1:
            relevant_vr = relevant_vr_candidates[0]
        else:
            if allow_multiple:
                relevant_vr = relevant_vr_candidates[0]
            else:
                e_msg = list()
                e_msg.append(f'There are more than one relevant variant '
                             f'records:')
                for vr in relevant_vr_candidates:
                    e_msg.append(str(vr))
                raise Exception('\n'.join(e_msg))

        return relevant_vr

    # main
    if vcf.index is None:
        raise NoVcfIndexError(f'Input vcf is not indexed.')

    matcher = matcher_exact

    if search_equivs:
        relevant_vr_candidates = search_equivalents(vcfspec, vcf, matcher, fasta)
    else:
        relevant_vr_candidates = search_identical(vcfspec, vcf, matcher)

    relevant_vr = get_relevant_vr(relevant_vr_candidates, allow_multiple)

    return relevant_vr


def fetch_relevant_vr_vcfpath(vcfspec, vcf_path, fasta=None, 
                              search_equivs=True):
    with pysam.VariantFile(vcf_path, 'r') as vcf:
        try:
            relevant_vr = fetch_relevant_vr(vcfspec, vcf, fasta = fasta, 
                                            search_equivs = search_equivs)
        except NoVcfIndexError:
            print(f'Input vcf path: {vcf_path}')
            raise

    return relevant_vr


# cosmic


# geneset gff3

def parse_transcript_tabixline(tabixline):
    annotitem = annotationdb.AnnotItem()

    attrs_parsed = parse_gff3_attrs(tabixline.attributes)
    annotitem['id'] = attrs_parsed['ID'].split(':')[1]

    ensembl_parser.set_feature_type(annotitem, 'transcript')
    annotitem['biotype'] = attrs_parsed['biotype']
    ensembl_parser.set_transcript_subtypes(annotitem, annotitem['biotype'], 
                                           annotitem['id'])

    if tabixline.strand == '+':
        annotitem['is_forward'] = True
    elif tabixline.strand == '-':
        annotitem['is_forward'] = False
    else:
        raise Exception(f'Unexpected transcript tabixline strand value: '
                        f'"{tabixline.strand}"')

    annotitem['chrom'] = tabixline.contig
    annotitem['start0'] = tabixline.start
    annotitem['start1'] = annotitem['start0'] + 1
    annotitem['end0'] = tabixline.end
    annotitem['end1'] = annotitem['end0']

    if 'Name' in attrs_parsed:
        annotitem['transcript_name'] = attrs_parsed['Name']
    else:
        annotitem['transcript_name'] = None

    annotitem['gene_id'] = attrs_parsed['Parent'].split(':')[1]

    return annotitem


def fetch_transcript(chrom, start0, end0, tabixfile_geneset):
    """Returns an empty AnnotItemDict object if the tabixfile does not 
    contain chrom.
    """

    transcript_dict = annotationdb.AnnotItemDict()
    if chrom in tabixfile_geneset.contigs:
        try:
            fetchresult = tabixfile_geneset.fetch(chrom, start0, end0)
        except Exception as e:
            raise Exception(f'tabixfile fetch failed: chrom: {chrom}, '
                            f'start0: {start0}, end0: {end0}') from e
            
        for tabixline in fetchresult:
            if check_tabixline_type_transcript(tabixline):
                annotitem = parse_transcript_tabixline(tabixline)
                transcript_dict[annotitem['id']] = annotitem

    return transcript_dict


def fetch_transcript_tabixline(chrom, start0, end0, transcript_id_list, 
                               tabixfile_geneset):
    if chrom not in tabixfile_geneset.contigs:
        raise Exception(f'chrom is not contained in the tabixfile.')

    candidates = list()
    fetched = tabixfile_geneset.fetch(chrom, start0, end0)
    for tabixline in fetched:
        mat = PAT_GENESET_ID.search(tabixline.attributes)
        if mat is not None:
            if mat.group('type') == 'transcript':
                if mat.group('id') in transcript_id_list:
                    candidates.append(tabixline)

    if len(candidates) != len(transcript_id_list):
        raise Exception(f'Unsuccessful transcript fetch: '
                        f'transcript_id_list: {transcript_id_list}, '
                        f'coords: ({chrom}, {start0}, {end0})')

    return candidates


def check_geneset_tabixline_type(tabixline, type):
    assert type in ('gene', 'transcript', 'CDS', 'exon', 'UTR'), (
        f'Allowed "type" values are: '
        f'"gene", "transcript", "CDS", "exon", "UTR"')

    if type == 'exon':
        return tabixline.feature == 'exon'
    elif type == 'UTR':
        return tabixline.feature in ('five_prime_UTR', 'three_prime_UTR')
    else:
        mat = PAT_GENESET_ID.search(tabixline.attributes)
        if mat is None:
            return False
        else:
            return (mat.group('type') == type)


def check_tabixline_type_transcript(tabixline):
    mat = PAT_GENESET_ID.search(tabixline.attributes)
    if mat is None:
        return False
    else:
        return (mat.group('type') == 'transcript')


def get_parent(tabixline):
    mat = PAT_GENESET_PARENT.search(tabixline.attributes)
    if mat is None:
        return None
    else:
        return mat.group('id')


# repeats bed

def parse_repeat_tabixline(tabixline):
    annotitem = annotationdb.AnnotItem()

    annotitem['chrom'] = tabixline.contig
    annotitem['start0'] = tabixline.start
    annotitem['end0'] = tabixline.end
    annotitem['start1'] = annotitem['start0'] + 1
    annotitem['end1'] = annotitem['end0']

    ensembl_parser.set_feature_type(annotitem, 'repeat')

    raw_split = tabixline.name.split(':')
    clsfml = raw_split[0].split('/')
    if len(clsfml) == 1:
        annotitem['class'] = clsfml[0]
        annotitem['family'] = clsfml[0]
    elif len(clsfml) == 2:
        annotitem['class'] = clsfml[0]
        annotitem['family'] = clsfml[1]
    else:
        raise Exception(f'Unexpected repeat name pattern: {tabixline.name}')
    annotitem['name'] = raw_split[1]

    annotitem['id'] = '_'.join([
        annotitem["name"], tabixline.contig, 
        str(tabixline.start), str(tabixline.end)])

    return annotitem


def fetch_repeat(chrom, start0, end0, tabixfile_repeats):
    """Returns an empty AnnotItemDict object if the tabixfile does not 
    contain chrom.
    """

    repeat_dict = annotationdb.AnnotItemDict()
    if chrom in tabixfile_repeats.contigs:
        try:
            fetchresult = tabixfile_repeats.fetch(chrom, start0, end0)
        except Exception as e:
            raise Exception(f'tabixfile fetch failed: chrom: {chrom}, '
                            f'start0: {start0}, end0: {end0}') from e

        for tabixline in fetchresult:
            annotitem = parse_repeat_tabixline(tabixline)
            repeat_dict[annotitem['id']] = annotitem

    return repeat_dict


# regulatory gff3

def parse_regulatory_tabixline(tabixline):
    annotitem = annotationdb.AnnotItem()

    annotitem['chrom'] = tabixline.contig
    annotitem['start0'] = tabixline.start
    annotitem['end0'] = tabixline.end
    annotitem['start1'] = annotitem['start0'] + 1
    annotitem['end1'] = annotitem['end0']

    ensembl_parser.set_feature_type(annotitem, 'regulatory')
    ensembl_parser.set_regulatory_subtypes(annotitem, tabixline.feature, 'vep')

    attrs_parsed = parse_gff3_attrs(tabixline.attributes)
    annotitem['id'] = attrs_parsed['id']
    annotitem['activity'] = dict(
        x.split(':') for x in attrs_parsed['activity'].split(','))

    annotitem['bound_start1'] = int(attrs_parsed['bound_start'])
    annotitem['bound_end1'] = int(attrs_parsed['bound_end'])
    annotitem['bound_start0'] = annotitem['bound_start1'] - 1
    annotitem['bound_end0'] = annotitem['bound_end1']

    return annotitem


def fetch_regulatory_tabixline(chrom, start0, end0, regulatory_id_list, 
                               tabixfile_regulatory):
    if chrom not in tabixfile_regulatory.contigs:
        raise Exception(f'chrom is not contained in the tabixfile.')

    candidates = list()
    fetched = tabixfile_regulatory.fetch(chrom, start0, end0)
    for tabixline in fetched:
        ID = get_ID_regulatory_tabixline(tabixline)
        if ID in regulatory_id_list:
            candidates.append(tabixline)

    if len(candidates) != len(regulatory_id_list):
        raise Exception(f'Unsuccessful regulatory fetch: '
                        f'regulatory_id_list: {regulatory_id_list}, '
                        f'coords: ({chrom}, {start0}, {end0})')

    return candidates


def fetch_regulatory(chrom, start0, end0, tabixfile_regulatory):
    """Returns an empty AnnotItemDict object if the tabixfile does not 
    contain chrom.
    """

    regulatory_dict = annotationdb.AnnotItemDict()
    if chrom in tabixfile_regulatory.contigs:
        try:
            fetchresult = tabixfile_regulatory.fetch(chrom, start0, end0)
        except Exception as e:
            raise Exception(f'tabixfile fetch failed: chrom: {chrom}, '
                            f'start0: {start0}, end0: {end0}') from e

        for tabixline in fetchresult:
            annotitem = parse_regulatory_tabixline(tabixline)
            regulatory_dict[annotitem['id']] = annotitem

    return regulatory_dict


def get_ID_regulatory_tabixline(tabixline):
    return PAT_REGULATORY_ID.search(tabixline.attributes).group(2)

# misc.

def parse_gff3_attrs(raw_string):
    return dict(x.split('=') for x in raw_string.split(';'))




import sys
import os
import re
import argparse
import subprocess
import time
import logging
import uuid
import textwrap
import itertools
import tempfile

import pyranges as pr

import handygenome
import handygenome.logutils as logutils
import handygenome.interval as libinterval
import handygenome.tools as tools
import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome
import handygenome.vcfeditor.misc as vcfmisc


SLURMBIN = '/usr/local/slurm/bin'
SQUEUE = os.path.join(SLURMBIN, 'squeue')
SBATCH = os.path.join(SLURMBIN, 'sbatch')
SCONTROL = os.path.join(SLURMBIN, 'scontrol')
SCANCEL = os.path.join(SLURMBIN, 'scancel')
SINFO = os.path.join(SLURMBIN, 'sinfo')

PBSBIN = '/usr/local/pbs/bin/'
PBSNODES = os.path.join(PBSBIN, 'pbsnodes')
QSTAT = os.path.join(PBSBIN, 'qstat')

RE_PATS = dict()
RE_PATS['scontrol_split_job'] = re.compile('^([^=]+)(=(.*))?$')
RE_PATS['scontrol_split_nodes'] = re.compile(r'(?<=\S)\s+(?=\S+=)')

DEFAULT_INTV_CHECK = 60  # seconds
DEFAULT_INTV_SUBMIT = 1  # seconds
DEFAULT_MAX_SUBMIT = None  # no limits

HELP_WIDTH = 50
DESCRIPTION_WIDTH = 80


# logging #

DEFAULT_DATEFMT = '%Z %Y-%m-%d %H:%M:%S' # KST 2022-03-23 22:12:34
DEFAULT_LOG_FORMATTERS = {
    'without_name': logging.Formatter(
        fmt='[%(asctime)s %(levelname)s] %(message)s', 
        datefmt=DEFAULT_DATEFMT,
    ),
    'with_name': logging.Formatter(
        fmt='[%(asctime)s %(levelname)s] %(name)s: %(message)s', 
        datefmt=DEFAULT_DATEFMT,
    ),
}


def get_logger(name=None, formatter=None, level='info', stderr=True, 
               filenames=None, append=False):
    if name is None:
        #name = str(uuid.uuid4())
        name = __name__.split('.')[-1]

    if formatter is None:
        formatter = DEFAULT_LOG_FORMATTERS['with_name']

    loglevel = getattr(logging, level.upper())

    logger = logging.getLogger(name)
    logger.setLevel(loglevel)
    logger.propagate = False

    if stderr:
        sh = logging.StreamHandler()
        sh.setLevel(loglevel)
        sh.setFormatter(formatter)
        logger.addHandler(sh)

    if filenames is not None:
        assert isinstance(filenames, (tuple, list))
        for fname in filenames:
            fh = logging.FileHandler(fname, mode=('a' if append else 'w'))
            fh.setLevel(loglevel)
            fh.setFormatter(formatter)
            logger.addHandler(fh)

    return logger


def get_debugging_logger(title=None, verbose=True):
    if title is None:
        title = '%(module)s.%(funcName)s'

    return get_logger(
        name=str(uuid.uuid4()),
        level=('debug' if verbose else 'info'),
        formatter=logging.Formatter(
            fmt=f'[%(asctime)s.%(msecs)03d {title}] line %(lineno)d: %(message)s', 
            datefmt='%Z %Y-%m-%d %H:%M:%S'
        ),
    )


# filesystem-related functions

def get_split_filenames(n_file, outdir, prefix, suffix):
    """
    Returns:
        A list of file names each of which looks like 
            <outdir>/<prefix>000<suffix>.
        File creation is not done.
    """
    return [
        os.path.join(outdir, prefix + idx_pad + suffix)
        for idx_pad in tools.get_padded_indices(n_file)
    ]


def check_infile_validity(infile):
    """
    Raise:
        If infile does not exist or user does not have read permission
    """

    if os.path.exists(infile):
        if os.path.isfile(infile):
            if not os.access(infile, os.R_OK):
                raise Exception(f'You do not have read permission on '
                                f'"{infile}".')
        else:
            raise Exception(f"'{infile}' is not a regular file.")
    else:
        raise Exception(f"'{infile}' does not exist.")


def check_indir_validity(indir):
    if os.path.exists(indir):
        if os.path.isdir(indir):
            if not os.access(indir, os.R_OK | os.X_OK):
                raise Exception(f'You do not have read and execution '
                                f'permission on "{indir}".')
        else:
            raise Exception(f"'{indir}' is not a directory.")
    else:
        raise Exception(f"'{indir}' does not exist.")


def check_outdir_validity(outdir):
    """
    Return: 
        True if outdir already exists, otherwise False.

    Raise:
        If 1) outdir is an existing non-directory file, or 
            2) outdir is an existing non-empty directory
    """

    if not os.path.exists(outdir):
        if os.access(os.path.dirname(outdir), os.W_OK | os.X_OK):
            return False
        else:
            raise Exception(f'You do not have w/x permission on '
                            f'"dirname" of "{outdir}".')

    elif os.path.isfile(outdir):
        raise Exception(f"'{outdir}' is an existing regular file.")

    elif os.path.isdir(outdir):
        if os.access(outdir, os.W_OK | os.X_OK):
            if len(os.listdir(outdir)) != 0:
                raise Exception(
                    f'"{outdir}" is an existing directory and is not empty.')
            else:
                return True
        else:
            raise Exception(
                f'"{outdir}" is an existing directory which you do not '
                f'have w/x permission on.')

    else:
        raise Exception(f'"{outdir}" is an existing file which is neither '
                        f'a regular file nor a directory.')


def check_outfile_validity(outfile, must_not_exist=False):
    """
    Return: 
        True if outfile already exists, otherwise False.
    Raise:
        permission problem
    """

    if not os.path.exists(outfile):
        if os.access(os.path.dirname(outfile), os.W_OK | os.X_OK):
            return False
        else:
            raise Exception(f'You do not have w/x permission on dirname '
                            f'of "{outfile}".')

    elif os.path.isfile(outfile):
        if must_not_exist:
            raise Exception(f'Specified file "{outfile}" must not exist '
                            f'in advance.')
        else:
            return True

    elif os.path.isdir(outfile):
        raise Exception(f"'{outfile}' is an existing directory.")

    else:
        raise Exception(f'"{outfile}" is an existing file which is neither '
                        f'a regular file nor a directory.')


def get_tmpfile_path(prefix=None, suffix=None, dir=None, where=None, delete=False, is_dir=False):
    """Args:
        where: alias for dir
    """
    # alias handling
    if where is not None:
        dir = where
    # main
    if dir is None:
        dir = os.getcwd()

    if delete:
        fd, path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
        os.close(fd)
        os.remove(path)
    else:
        if is_dir:
            path = tempfile.mkdtemp(prefix=prefix, suffix=suffix, dir=dir)
        else:
            fd, path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
            os.close(fd)

    return path


def get_tmpdir_paths(subdirs, prefix=None, suffix=None, where=os.getcwd(), 
                     top_path=None, root_path=None):
    """Args:
        subdirs: An iterable which contains subdirectory basenames
    """

    root_name = 'root'
    assert root_name not in subdirs, (
        f'"subdirs" argument must not include the name "{root_name}".')

    # alias handling
    if (top_path is not None) and (root_path is None):
        root_path = top_path

    if root_path is None:
        root_path = os.path.abspath(
            get_tmpfile_path(prefix=prefix, suffix=suffix, where=where, 
                             delete=False, is_dir=True))
    else:
        exists = check_outdir_validity(root_path)
        if not exists:
            os.mkdir(root_path)

    tmpdir_paths = dict()
    tmpdir_paths[root_name] = root_path
    for basename in subdirs:
        tmpdir_paths[basename] = os.path.join(tmpdir_paths[root_name], 
                                              basename)
        os.mkdir(tmpdir_paths[basename])

    return tmpdir_paths


# ARGPARSE SETUP FUNCTIONS

class CustomFormatter(
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.RawTextHelpFormatter):
    pass


def init_parser(description=None):
    if description is None:
        parser = argparse.ArgumentParser(
            description=None,
            formatter_class=CustomFormatter, add_help=False)
    else:
        parser = argparse.ArgumentParser(
            description=textwrap.fill(description, width=DESCRIPTION_WIDTH),
            formatter_class=CustomFormatter, add_help=False)

    required = parser.add_argument_group(
        title='REQUIRED', 
        description=textwrap.fill(
            'Required ones which accept 1 or more arguments.',
            width=DESCRIPTION_WIDTH))

    optional = parser.add_argument_group(
        title='OPTIONAL', 
        description=textwrap.fill(
            'Optional ones which accept 1 or more arguments.',
            width=DESCRIPTION_WIDTH))

    flag = parser.add_argument_group(
        title='FLAG', 
        description=textwrap.fill(
            'Optional ones which accept 0 argument.',
            width=DESCRIPTION_WIDTH))

    add_help_arg(flag)

    parser_dict = {'main': parser, 'required': required, 
                   'optional': optional, 'flag': flag}

    return parser_dict


def get_basic_parser():
    parser_dict = init_parser()

    add_infile_arg(parser_dict['required'], required=True)
    add_outfile_arg(parser_dict['required'], required=True, 
                    must_not_exist=True)
    add_fasta_arg(parser_dict['required'], required=True)

    add_refver_arg(parser_dict['optional'], required=False, choices='all')
    add_outfmt_arg(parser_dict['optional'], required=False)
    add_scheduler_args(parser_dict['optional'], default_parallel=1,
                       default_sched='slurm',
                       default_check_interval=DEFAULT_INTV_CHECK,
                       default_submit_interval=DEFAULT_INTV_SUBMIT)

    return parser_dict['main']


def add_infile_arg(parser, required=True, help=f'Input vcf file path.'):
    parser.add_argument(
        '-i', '--infile', dest='infile_path', required=required, 
        type=arghandler_infile, metavar='<input file path>', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_infilelist_arg(
        parser, required=True, 
        help=f'One or more input file paths, separated with whitespaces.'):
    parser.add_argument(
        '--infilelist', dest='infile_path_list', required=required, 
        nargs='+', type=arghandler_infile, metavar='<input file path>', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_indir_arg(parser, required=True, help=f'Input directory path.'):
    parser.add_argument(
        '--dir', dest='indir_path', required=required, type=arghandler_indir,
        metavar='<input directory path>',
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_outfile_arg(
    parser, required=True, must_not_exist='ask',
    help='Output vcf file path. Must not exist in advance.',
):
    handler = get_arghandler_outfile(must_not_exist)
    parser.add_argument(
        '-o', '--outfile', dest='outfile_path', required=required, 
        type=handler, metavar='<output file path>',
        help=textwrap.fill(help, width=HELP_WIDTH),
    )


def add_outdir_arg(
        parser, required=True,
        help=('Output directory path. It will be created if it does not '
              'exist; otherwise it must be an empty directory.')):
    parser.add_argument(
        '--outdir', dest='outdir_path', required=True,
        type=arghandler_outdir, metavar='<output directory>',
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_fasta_arg(parser, required=True, help=None):
    available_versions = ", ".join(refgenome.REFVERINFO.list_known_refvers())

    if help is None:
        help=(f'A fasta file path or, alternatively, a reference genome '
              f'version (one of {available_versions}), '
              f'in which case a preset fasta file for the reference version '
              f'is used.')

    parser.add_argument(
        '-f', '--fasta', dest='fasta_path', required=required,
        type=arghandler_fasta, metavar='<fasta path>', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_refver_arg(parser, required=True, choices='all', help=None):
    if choices == 'all':
        allowed_vals = refgenome.RefverDict.known_refvers
    elif choices == 'mouse':
        allowed_vals = ('mm9', 'mm10', 'mm39')
    elif choices == 'human':
        allowed_vals = ('hg18', 'hg19', 'hg38')
    else:
        allowed_vals = choices

    if help is None:
        help = f'Reference genome version. Must be one of {allowed_vals}.'

    parser.add_argument(
        '--refver', dest='refver', required=required, default=None, 
        choices=allowed_vals, metavar='<reference genome version>', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_outfmt_arg(
        parser, required=False, default='z',
        help=(f'Output vcf file format (bcftools style). Must be one of: '
              f'v (uncompressed VCF), z (compressed VCF), '
              f'u (uncompressed BCF), b (compressed BCF).')):
    assert default in ('v', 'z', 'u', 'b')
    parser.add_argument(
        '-O', '--output-format', dest='mode_pysam', required=required, 
        default=default, choices=('v', 'z', 'u', 'b'), 
        type=(lambda x: vcfmisc.PYSAM_MODE_DICT[x]),
        metavar='<output format>',
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_parallel_arg(parser, required=False, default=1,
                     help=f'Number of parallelization.'):
    parser.add_argument(
        '-p', '--parallel', dest='parallel', required=required,
        default=default, type=int,
        metavar='<number of parallelization>', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_sched_arg(
        parser, required=False, default='slurm',
        help=f'Parallelization method. Must be "slurm" or "local".'):
    assert default in ('slurm', 'local')
    parser.add_argument(
        '--sched', dest='sched', required=required, default=default, 
        choices=('slurm', 'local'), metavar='<"slurm" or "local">', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_checkint_arg(
        parser, required=False, default=DEFAULT_INTV_CHECK,
        help=f'Slurm job status check interval in seconds.'):
    parser.add_argument(
        '--slurm-check-interval', dest='intv_check', required=required,
        default=default, type=float, metavar='<slurm check interval>', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_submitint_arg(
        parser, required=False, default=DEFAULT_INTV_SUBMIT,
        help=f'Slurm job submission interval in seconds.'):
    parser.add_argument(
        '--slurm-submit-interval', dest='intv_submit', required=required,
        default=default, type=float, metavar='<slurm submit interval>', 
        help=textwrap.fill(help, width=HELP_WIDTH))


def add_maxsubmit_arg(
    parser, 
    required=False, 
    default=DEFAULT_MAX_SUBMIT,
    help=f'Allowed maximum number of pending and running jobs at a time.',
):
    parser.add_argument(
        '--slurm-max-submit', dest='max_submit', required=required,
        default=default, type=int, metavar='<slurm maximum submit>', 
        help=textwrap.fill(help, width=HELP_WIDTH)
    )


def add_index_arg(parser_dict,
                  help='If set, output vcf file is not indexed.'):
    parser_dict['flag'].add_argument(
        '--donot-index', dest='donot_index', action='store_true',
        help=textwrap.fill(help, width=HELP_WIDTH))

#######

def add_incl_region_arg(
        parser, required=False,
        help=(f'Bed file (gzipped or not) representing regions to be included.'
              f'If not set, whole genomic regions are used.')):
    parser.add_argument('--included-region-bed', dest='incl_bed_path', 
                        required=required, help=help)


def add_excl_region_arg(
        parser, required=False,
        help=(f'Bed file (gzipped or not) representing regions to be excluded.'
              f'If not set, no region is excluded.')):
    parser.add_argument('--excluded-region-bed', dest='excl_bed_path', 
                        required=required, help=help)


def get_incl_gr_from_args(args):
    if args.incl_region_path is None:
        if args.fasta_path is not None:
            chromdict = refgenome.ChromDict.from_fasta_path(args.fasta_path)
        elif args.refver is not None:
            chromdict = refgenome.ChromDict.from_refver(refver)
        else:
            raise Exception(
                f'"incl_region_path", "fasta_path", "refver" are all not set.'
            )

        #gr_incl = chromdict.to_interval_list().to_gr()
        gr_incl = chromdict.to_gr(assembled_only=False, as_gr=True)
    else:
        gr_incl = pr.read_bed(args.incl_region_path)

    return gr_incl


def get_merged_intervals_from_args(args):
    gr_incl = get_incl_gr_from_args(args)
    if args.excl_region_path is None:
        gr_merged = gr_incl
    else:
        gr_excl = pr.read_bed(args.excl_region_path)
        gr_merged = gr_incl.subtract(gr_excl)

    return libinterval.IntervalList.from_gr(gr_merged)
            

########

def add_help_arg(parser, help='show this help message and exit'):
    parser.add_argument('-h', '--help', action='help',
                        help=textwrap.fill(help, width=HELP_WIDTH))


# functions which receive 'parser_dict' argument

def add_logging_args(parser_dict):
    parser_dict['optional'].add_argument(
        '--log', required=False, 
        help=textwrap.fill(
            f'If used, progress messages will be written to this file. '
            f'Existing file will be truncated.',
            width=HELP_WIDTH))

    parser_dict['flag'].add_argument(
        '--append-log', dest='append_log', action='store_true',
        help=textwrap.fill('If set, existing log file is not trucated.',
                           width=HELP_WIDTH))

    parser_dict['flag'].add_argument(
        '--silent', action='store_true',
        help=textwrap.fill(
            'If set, progress messages will not be printed to the terminal.',
            width=HELP_WIDTH))


def add_scheduler_args(
    parser_dict, 
    default_parallel=1, 
    default_sched='slurm',
    default_check_interval=DEFAULT_INTV_CHECK,
    default_submit_interval=DEFAULT_INTV_SUBMIT,
    default_max_submit=DEFAULT_MAX_SUBMIT,
):
    add_parallel_arg(parser_dict['optional'], required=False, default=default_parallel)
    add_sched_arg(parser_dict['optional'], required=False, default=default_sched)
    add_checkint_arg(parser_dict['optional'], required=False, default=default_check_interval)
    add_submitint_arg(parser_dict['optional'], required=False, default=default_submit_interval)
    add_maxsubmit_arg(parser_dict['optional'], required=False, default=default_max_submit)


def add_rmtmp_arg(parser_dict):
    parser_dict['flag'].add_argument(
        '--donot-remove-tmp', dest='donot_rm_tmp', action='store_true',
        help=textwrap.fill(
            'If set, temporary files are not removed (which are removed by '
            'default).',
            width=HELP_WIDTH))


# arghandlers

def arghandler_infile(arg):
    arg = os.path.abspath(arg)

    try:
        check_infile_validity(arg)
    except Exception as e:
        raise argparse.ArgumentTypeError(str(e))
    else:
        return arg


def arghandler_infilelist(arg):
    """Args:
        arg: A list of input file paths 
    """

    return [arghandler_infile(x) for x in arg]


def arghandler_indir(arg):
    arg = os.path.abspath(arg)

    try:
        check_indir_validity(arg)
    except Exception as e:
        raise argparse.ArgumentTypeError(str(e))
    else:
        return arg


def arghandler_outfile_ask(arg):
    arg = os.path.abspath(arg)

    try:
        exists = check_outfile_validity(arg, must_not_exist=False)
    except Exception as e:
        raise argparse.ArgumentTypeError(str(e))
    else:
        if exists:
            msg = (f'Specified output file "{arg}" already exists. Would '
                   f'you like to proceed with overwriting? (y/n) ')
            ans = input(msg)
            while True:
                if ans == 'y':
                    return arg
                elif ans == 'n':
                    sys.exit(1)
                else:
                    ans = input('Please enter "y" or "n". ')
                    continue
        else:
            return arg


def arghandler_outfile_mayexist(arg):
    arg = os.path.abspath(arg)

    try:
        check_outfile_validity(arg, must_not_exist=False)
    except Exception as e:
        raise argparse.ArgumentTypeError(str(e))
    else:
        return arg


def arghandler_outfile_mustbeabsent(arg):
    arg = os.path.abspath(arg)

    try:
        check_outfile_validity(arg, must_not_exist=True)
    except Exception as e:
        raise argparse.ArgumentTypeError(str(e))
    else:
        return arg


def get_arghandler_outfile(must_not_exist):
    assert must_not_exist in ('yes', 'no', 'ask'), (
        '"must_not_exist" argument must be on of "yes", "no", or "ask".'
    )
    if must_not_exist == 'yes':
        return arghandler_outfile_mustbeabsent
    elif must_not_exist == 'no':
        return arghandler_outfile_mayexist
    elif must_not_exist == 'ask':
        return arghandler_outfile_ask


def arghandler_outdir(arg):
    arg = os.path.abspath(arg)

    try:
        exists = check_outdir_validity(arg)
    except Exception as e:
        raise argparse.ArgumentTypeError(str(e))
    else:
        if not exists:
            os.mkdir(arg)
        return arg


def arghandler_fasta(arg):
    if arg in refgenome.REFVERINFO.list_known_refvers():
    #if arg in common.DEFAULT_FASTA_PATHS.get_valid_keys():
        return refgenome.get_fasta_path(arg)
    else:
        return arghandler_infile(arg)


# decorators

def get_deco_arghandler_infile(argname):
    def decorator(func):
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.arguments[argname] = arghandler_infile(ba.arguments[argname])

            return func(*ba.args, **ba.kwargs)

        return wrapper

    return decorator


def get_deco_arghandler_infilelist(argname):
    def decorator(func):
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.arguments[argname] = arghandler_infilelist(
                ba.arguments[argname])

            return func(*ba.args, **ba.kwargs)

        return wrapper

    return decorator


##########################################################


# SLURM JOB HANDLING

class Job:
    """Designed for slurm

    Attributes:
        jobscript_path: Path of job script file to run.
        jobscript_string: Multi-line string to be written to stdin of sbatch.
        jobid: JobId as integer. None before job submission.
        submitted: True if job has been submitted, False otherwise.
        jobstate
        exitcode: type int
        success: 
            None: 
                1) before job has been finished
                2) job has been finished but its state cannot be checked 
                    because job id has disappeared from slurm
            True if the job has finished successfully
            False if the job has finished not successfully
        no_jobid: 
            None: before job has been submitted
            True: job id is no longer remaining on slurm
            False: job id exists on slurm
        scontrol_result: Dictionary which contains result of 
            "scontrol show job <jobid>" command.
        sbatch_err_info: An informative string created when sbatch returns 
            a nonzero exit code.
        stdout_path: job stdout path
        stderr_path: job stderr path
        status
        logger
    """
    jobstates_pending = [
        'CONFIGURING', 'PENDING',
    ]
    jobstates_running = [
        'COMPLETING', 'RUNNING', 'RESV_DEL_HOLD', 
        'REQUEUE_FED', 'REQUEUE_HOLD', 'RESIZING', 
        'SIGNALING', 'SPECIAL_EXIT', 'STAGE_OUT', 'STOPPED', 
        'SUSPENDED',
    ]
    jobstates_finished = [
        'CANCELLED', 'COMPLETED', 'BOOT_FAIL', 'DEADLINE', 
        'FAILED', 'NODE_FAIL', 'OUT_OF_MEMORY', 'PREEMPTED', 
        'TIMEOUT',
    ]
    jobstates_unknown = [
        'REVOKED',
    ]

    @deco.get_deco_num_set_differently(('jobscript_path', 'jobscript_string'), 1)
    def __init__(self, jobscript_path=None, jobscript_string=None, 
                 verbose=True, logger=None):
        self.jobscript_path = jobscript_path
        self.jobscript_string = jobscript_string
        self.jobid = None
        self.submitted = False
        self.jobstate = None
        self.exitcode = None
        self.success = None
        self.no_jobid = None
        self.scontrol_result = None
        self.sbatch_err_info = None
        self.stdout_path = None
        self.stderr_path = None
        self.status = {
            'pending': False, 
            'running': False, 
            'finished': False,
        }

        if logger is None:
            self.logger = get_logger(
                formatter=DEFAULT_LOG_FORMATTERS['without_name'], 
                stderr=verbose)
        else:
            self.logger = logger

    def __repr__(self):
        infostr = ', '.join(
            f'{key}: {getattr(self, key)}'
            for key in ('jobid', 'jobstate', 'exitcode', 'jobscript_path')
        )
        return f'<Job ({infostr})>'

    def submit(self):
        if self.jobscript_path is None:
            self._submit_string()
        else:
            self._submit_path()

    def cancel(self):
        assert self.submitted, f'The job is not yet submitted.'
        p = subprocess.run([SCANCEL, str(self.jobid)],
                           capture_output=True, text=True)
        if p.returncode != 0:
            self.scancel_err_info = textwrap.dedent(f"""\
                Job cancellation using scancel failed.

                job id: {self.jobid}
                scancel stdout: {p.stdout}
                scancel stderr: {p.stderr}
                scancel exit code: {p.returncode}""")
            raise Exception(self.scancel_err_info)
        else:
            self.update()
            self.logger.info(f'Cancelled a job: JobID - {self.jobid}')

    def set_status(self, key):
        self.status[key] = True
        other_keys = set(self.status.keys()).difference([key])
        for x in other_keys:
            self.status[x] = False

    def update(self):
        assert self.submitted

        scontrol_result, no_jobid = get_scontrol_job_result(self.jobid)[:2]
        self.no_jobid = no_jobid

        if self.no_jobid:
            # self.success and self.scontrol_result are not touched
            self.set_status('finished')
        else:
            self.scontrol_result = scontrol_result

            # update self.stderr/stdout path
            self.stdout_path = self.scontrol_result['StdOut']
            self.stderr_path = self.scontrol_result['StdErr']

            # update self.jobstate and self.status
            self.jobstate = self.scontrol_result['JobState']
            if self.jobstate in self.__class__.jobstates_pending:
                self.set_status('pending')
            elif self.jobstate in self.__class__.jobstates_running:
                self.set_status('running')
            elif self.jobstate in self.__class__.jobstates_finished:
                self.set_status('finished')
            else:
                raise Exception(textwrap.dedent(f"""\
                    Unable to interpret JobState.
                    JobState: {self.jobstate}
                    scontrol result: {self.scontrol_result}"""))

            # update self.exitcode, self.success
            if self.status['finished']:
                self.success = (self.jobstate == 'COMPLETED')
                self.exitcode = int(scontrol_result['ExitCode'].split(':')[0])

    ##########################

    def _submit_string(self):
        p = subprocess.run([SBATCH], input=self.jobscript_string,
                           capture_output=True, text=True)

        if p.returncode != 0:
            self.sbatch_err_info = textwrap.dedent(f"""\
                Job submission using sbatch failed.

                string written to the stdin of sbatch: {self.jobscript_string}
                sbatch stdout: {p.stdout}
                sbatch stderr: {p.stderr}
                sbatch exit code: {p.returncode}""")
            raise Exception(self.sbatch_err_info)
        else:
            self.jobid = int(p.stdout.split()[-1])
            self.submitted = True
            self.logger.info(f'Submitted a job: JobID - {self.jobid}')

    def _submit_path(self):
        p = subprocess.run([SBATCH, self.jobscript_path],
                           capture_output=True, text=True)

        if p.returncode != 0:
            self.sbatch_err_info = textwrap.dedent(f"""\
                Job submission using sbatch failed.

                job script path: {self.jobscript_path}
                sbatch stdout: {p.stdout}
                sbatch stderr: {p.stderr}
                sbatch exit code: {p.returncode}""")
            raise Exception(self.sbatch_err_info)
        else:
            self.jobid = int(p.stdout.split()[-1])
            self.submitted = True
            self.logger.info(f'Submitted a job: JobID - {self.jobid}')


class JobList(list):
    """
    Attributes:
        logger
        success
        
    """

    def __init__(
        self, 
        jobscript_path_list, 
        intv_submit=DEFAULT_INTV_SUBMIT,
        intv_check=DEFAULT_INTV_CHECK,
        max_submit=DEFAULT_MAX_SUBMIT,
        verbose=True, 
        logpath=None, 
        logger=None,
        job_status_logpath=None,
    ):
        # general logger
        self.verbose = verbose
        if logger is None:
            self.logger = get_logger(
                formatter=DEFAULT_LOG_FORMATTERS['without_name'], 
                stderr=verbose, 
                filenames=(
                    None if logpath is None else [logpath]
                ),
            )
        else:
            self.logger = logger

        # job status summary logger
        self.job_status_logpath = job_status_logpath

        # other attrs
        for jobscript_path in jobscript_path_list:
            self.append(
                Job(jobscript_path=jobscript_path, verbose=verbose, logger=self.logger)
            )

        self.intv_submit = intv_submit
        self.intv_check = intv_check
        self.max_submit = max_submit
        self.success = None
        self.sublists = dict()
        self.update()

    def write_job_status_log(self, msg):
        if self.job_status_logpath is not None:
            #with open(self.job_status_logpath, 'wt') as f:
            with tools.openfile(self.job_status_logpath, 'a') as f:
                f.write(f'[{logutils.get_timestamp()}] {msg}')

    def get_num_pending_running(self):
        return len(self.sublists['pending']) + len(self.sublists['running'])

    def submit(self):
        if self.max_submit is None:
            num_to_submit = len(self.sublists['notsubmit'])
        else:
            num_pending_running = self.get_num_pending_running()
            num_to_submit = min(
                self.max_submit - num_pending_running, 
                len(self.sublists['notsubmit']),
            )

        if num_to_submit > 0:
            for job in self.sublists['notsubmit'][:num_to_submit]:
                job.submit()
                time.sleep(self.intv_submit)
            self.update()

    def cancel_all_running(self):
        for job in self.sublists['running']:
            job.cancel()


    def make_infostring(self, sublist_key):
        n_jobs = len(self.sublists[sublist_key])
        details = ', '.join(
            job.__repr__()  # 'jobid', 'jobstate', 'exitcode'
            for job in self.sublists[sublist_key]
        )
        return f'{n_jobs} ({details})'

    def log_status(self):
        n_notsubmit = len(self.sublists['notsubmit'])
        n_pending = len(self.sublists['pending'])
        n_running = len(self.sublists['running'])
        n_finished = len(self.sublists['finished'])

        info_pending = self.make_infostring('pending')
        info_running = self.make_infostring('running')
        info_finished = self.make_infostring('finished')

        # more verbose information
        self.write_job_status_log(
            textwrap.dedent(
                f"""\
                Current job status:
                    Not submitted yet: {n_notsubmit}

                    Pending: {info_pending}

                    Running: {info_running}

                    Finished: {info_finished}
                """
            )
        )

        # concise information (to be printed to stderr)
        self.logger.info(
            textwrap.dedent(
                f"""\
                Current job status:
                    Not submitted yet: {n_notsubmit}
                    Pending: {n_pending}
                    Running: {n_running}
                    Finished: {n_finished}"""
            )
        )

    def log_epilogue(self):
        n_success = len(self.sublists['success'])
        n_failure = len(self.sublists['failure'])
        n_unknown = len(self.sublists['unknown'])

        info_success = self.make_infostring('success')
        info_failure = self.make_infostring('failure')
        info_unknown = self.make_infostring('unknown')

        self.write_job_status_log(
            textwrap.dedent(
                f"""\
                All finished.
                    Successful jobs: {info_success}

                    Failed jobs: {info_failure}

                    Jobs with unknown exit statuses: {info_unknown}
                """
            )
        )
        self.logger.info(
            textwrap.dedent(
                f"""\
                All finished.
                    Successful jobs: {n_success}
                    Failed jobs: {n_failure}
                    Jobs with unknown exit statuses: {n_unknown}"""
            )
        )

    def submit_and_wait(self):
        # main
        try:
            while True:
                self.update()
                self.submit()
                self.log_status()
                if all(job.status['finished'] for job in self):
                    break
                else:
                    time.sleep(self.intv_check)
                    continue
        except KeyboardInterrupt:
            self.logger.info(
                f'RECEIVED A KEYBOARD INTERRUPT; '
                f'will cancel all pending and running jobs with scancel,'
                f' then exit immediately.'
            )
            for job in itertools.chain(
                self.sublists['pending'], self.sublists['running']
            ):
                job.cancel()
            raise SystemExit(1)
        else:
            self.set_sublists()
            self.log_epilogue()
            self.success = all(job.success for job in self)

    def recover_cancelled_jobs(self):
        cancelled_jobs = [
            (idx, job) for (idx, job) in enumerate(self) if job.jobstate == 'CANCELLED'
        ]
        for idx, job in cancelled_jobs:
            del self[idx]

        for idx, job in cancelled_jobs:
            self.append(
                Job(
                    jobscript_path=job.jobscript_path, 
                    verbose=self.verbose, 
                    logger=self.logger,
                )
            )

    def run(self):
        while True:
            self.submit_and_wait()
            if any((job.jobstate == 'CANCELLED') for job in self):
                self.recover_cancelled_jobs()
                continue
            else:
                break

    def get_exitcodes(self):
        return [job.exitcode for job in self]

    def update(self):
        self.update_jobs()
        self.set_sublists()

    def update_jobs(self):
        for job in self:
            if job.submitted:
                job.update()

    def set_sublists(self):
        for key in (
            'notsubmit', 
            'pending', 
            'running', 
            'finished', 
            'success', 
            'failure', 
            'unknown',
        ):
            self.sublists[key] = list()

        for job in self:
            if not job.submitted:
                self.sublists['notsubmit'].append(job)
            if job.status['pending']:
                self.sublists['pending'].append(job)
            if job.status['running']:
                self.sublists['running'].append(job)
            if job.status['finished']:
                self.sublists['finished'].append(job)

            if job.success is True:
                self.sublists['success'].append(job)
            if job.success is False:
                self.sublists['failure'].append(job)
            if job.success is None:
                self.sublists['unknown'].append(job)

#    def set_sublists_statuses(self):
#        self.sublists['notsubmit'] = [job for job in self 
#                                      if not job.submitted]
#        self.sublists['pending'] = [job for job in self 
#                                    if job.status['pending']]
#        self.sublists['running'] = [job for job in self 
#                                    if job.status['running']]
#        self.sublists['finished'] = [job for job in self 
#                                     if job.status['finished']]
#
#    def set_sublists_exitcodes(self):
#        self.sublists['success'] = [job for job in self 
#                                    if job.success is True]
#        self.sublists['failure'] = [job for job in self 
#                                    if job.success is False]
#        self.sublists['unknown'] = [job for job in self 
#                                    if job.success is None]


def get_scontrol_job_result(jobid):
    """
    Returns:
        A tuple (scontrol_result, no_jobid, returncode, stderr)

        scontrol_result: A dict with keys: 'JobId', 'JobName', 'UserId', 
            'GroupId', 'MCS_label', 'Priority', 'Nice', 'Account', 'QOS', 
            'JobState', 'Reason', 'Dependency', 'Requeue', 'Restarts', 
            'BatchFlag', 'Reboot', 'ExitCode', 'RunTime', 'TimeLimit', 
            'TimeMin', 'SubmitTime', 'EligibleTime', 'AccrueTime', 
            'StartTime', 'EndTime', 'Deadline', 'SuspendTime', 
            'SecsPreSuspend', 'LastSchedEval', 'Partition', 'AllocNode:Sid', 
            'ReqNodeList', 'ExcNodeList', 'NodeList', 'BatchHost', 'NumNodes', 
            'NumCPUs', 'NumTasks', 'CPUs/Task', 'ReqB:S:C:T', 'TRES', 
            'Socks/Node', 'NtasksPerN:B:S:C', 'CoreSpec', 'MinCPUsNode', 
            'MinMemoryNode', 'MinTmpDiskNode', 'Features', 'DelayBoot', 
            'OverSubscribe', 'Contiguous', 'Licenses', 'Network', 'Command', 
            'WorkDir', 'StdErr', 'StdIn', 'StdOut', 'Power', 'NtasksPerTRES:0'
        no_jobid: True or False 
        returncode
        stderr
    """

    cmdargs = [ SCONTROL, '-o', 'show', 'job', str(jobid) ]
    p = subprocess.run(args=cmdargs, text=True, capture_output=True)

    returncode = p.returncode
    stderr = p.stderr

    if p.returncode != 0:
        if p.stderr.strip() == 'slurm_load_jobs error: Invalid job id specified':
            scontrol_result = None
            no_jobid = True
        else:
            e_msg = f'''\
scontrol show job finished with nonzero exit code.
commands: {cmdargs}
stdout: {p.stdout}
stderr: {p.stderr}
exit code: {p.returncode}'''
            raise Exception(e_msg)

    else:
        no_jobid = False

        scontrol_result = dict()
        for word in p.stdout.split():
            wordsp = word.split('=', maxsplit = 1)
            if len(wordsp) == 1:
                key = wordsp[0]
                val = None
            elif len(wordsp) == 2:
                key, val = wordsp
                """
            else:
                e_msg = f'''\
Field with more than 2 "=" character found from "scontrol show job" output.
Field: {word}
scontrol command: {cmdargs}'''
                raise Exception(e_msg)
                """
    
            scontrol_result[key] = val
    
    return scontrol_result, no_jobid, returncode, stderr


###############################


def run_jobs(
    jobscript_paths, 
    sched, 
    intv_check, 
    intv_submit, 
    max_submit, 
    logger, 
    log_dir, 
    job_status_logpath=None,
    raise_on_failure=True,
):
    assert sched in ('local', 'slurm')

    if sched == 'local':
        success, exitcode_list = run_jobs_local(jobscript_paths)
    elif sched == 'slurm':
        joblist = JobList(
            jobscript_paths, 
            intv_check=intv_check,
            intv_submit=intv_submit,
            max_submit=max_submit,
            logger=logger,
            job_status_logpath=job_status_logpath,
        )
        joblist.run()

        success = joblist.success
        exitcode_list = joblist.get_exitcodes()

    if raise_on_failure:
        if not success:
            raise SystemExit(
                (f'One or more jobs have finished unsuccessfully. Refer to '
                 f'log files in {log_dir}'))
        else:
            return None
    else:
        return success, exitcode_list


def run_jobs_local(jobscript_path_list):
    plist = list()
    for jobscript_path in jobscript_path_list:
        p = subprocess.Popen([jobscript_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        plist.append(p)

    for p in plist:
        p.wait()

    exitcode_list = [p.returncode for p in plist]
    success = all(x == 0 for x in exitcode_list)

    return success, exitcode_list


"""
JOB SCRIPT GENERATION
"""

def make_multiline_command(lines, leading_taps = 0):
    new_lines = list()
    new_lines.append( '\t'*leading_taps + lines[0] + ' \\' )
    new_lines.extend( [ '\t'*(leading_taps+1) + x + ' \\' for x in lines[1:-1] ] )
    new_lines.append( '\t'*(leading_taps+1) + lines[-1] )

    return '\n'.join(new_lines)


def make_jobscript_string(lines, shell = False, python = False, **kwargs):
    string_list = list()


    string_list.append('#!' + handygenome.OPTION['bash'])

    if 'N' not in kwargs:
        kwargs['N'] = 1
    if 'n' not in kwargs:
        kwargs['n'] = 1
    if 'o' not in kwargs:
        kwargs['o'] = '/dev/null'
    if 'c' not in kwargs:
        kwargs['c'] = 1

    for k,v in kwargs.items():
        if v is not None:
            string_list.append(f'#SBATCH -{k} {v}')

    string_list.append('')

    for line in lines:
        string_list.append(line)

    return '\n'.join(string_list)


def make_jobscript(jobscript_path, lines, **kwargs):
    check_outfile_validity(jobscript_path)
    jobscript_string = make_jobscript_string(lines, **kwargs)
    with open(jobscript_path, 'w') as outfile:
        outfile.write(jobscript_string)




import re
import os
import gzip
import multiprocessing
#import shutil

import pysam
import Bio.SeqIO
import Bio.SeqUtils
import numpy as np

import handygenome.tools as tools
import handygenome.logutils as logutils


def make_index(fasta_path):
    """If input file is bgzipped, *.fai and *.gzi files are created."""
    #fai_path = fasta_path + '.fai'
    #if not os.path.exists(fai_path):
    _ = pysam.faidx(fasta_path)


##################
# fasta renaming #
##################

def _chrom_converter_handler(chrom_converter):
    if isinstance(chrom_converter, dict):
        def converter(x):
            if x in chrom_converter:
                return chrom_converter[x]
            else:
                return None
        return converter
    elif callable(chrom_converter):
        return chrom_converter
    else:
        raise Exception(f'Invalid input type')


def rename_fasta(infile_path, outfile_path, chrom_converter):
    chrom_converter = _chrom_converter_handler(chrom_converter)

    with (
        tools.openfile(infile_path, 'rt') as infile,
        open(outfile_path, 'wt') as outfile,
    ):
        def record_iter():
            for record in Bio.SeqIO.parse(infile, 'fasta'):
                new_id = chrom_converter(record.id)
                if new_id is None:
                    continue

                record.id = new_id
                record.name = None
                record.description = ''
                yield record

        Bio.SeqIO.write(record_iter(), outfile, 'fasta')


def rename_fasta_compress(infile_path, outfile_path, chrom_converter, index=True):
    """Not to be used since fetching from bgzipped fasta is much slower than
    from plain text fasta. (1.68 ms vs 20.5 ms)
    """
    # make target functions
    def pipe_out(pipe_path, outfile_path, index):
        """Read from fifo, bgzip-compress, and write to final output file"""
        pysam.tabix_compress(pipe_path, outfile_path, force=True)
        if index:
            make_index(outfile_path)

    # make fifo
    pipe_path = tools.get_tmpfile_path(prefix=f'{infile_path}_', suffix=f'.fifo', delete=True)
    os.mkfifo(pipe_path)

    # run
    subp1 = multiprocessing.Process(
        target=rename_fasta_nocompress, 
        args=(infile_path, pipe_path, chrom_converter),
    )
    subp2 = multiprocessing.Process(
        target=pipe_out, 
        args=(pipe_path, outfile_path, index),
    )
    subp1.start()
    subp2.start()
    subp1.join()
    subp2.join()

    # remove fifo
    os.remove(pipe_path)

    # raise if erroneous
    if (subp1.exitcode != 0) or (subp2.exitcode != 0):
        raise Exception(f'Finished with an error.')




############
# N region #
############

def make_N_region_coords(fasta):
    pat = re.compile('N+')
    chroms = list()
    start0s = list()
    end0s = list()
    for chrom in fasta.references:
        seq = fasta.fetch(chrom)
        offset = 0
        while True:
            mat = pat.search(seq)
            if mat is None:
                break
            else:
                span = mat.span()
                chroms.append(chrom)
                start0s.append(offset + span[0])
                end0s.append(offset + span[1])

                offset += span[1]
                seq = seq[span[1]:]

    return chroms, start0s, end0s


###############
# GC fraction #
###############

def calculate_gc_fractions_old(chroms, start0s, end0s, fasta, window=None, as_array=True):
    logutils.log('Beginning gc fraction calculation')
    get_fetchargs = _make_get_fetchargs(window, fasta)
    val_gen = (
        float(Bio.SeqUtils.gc_fraction(fasta.fetch(*get_fetchargs(chrom, start0, end0))))
        for chrom, start0, end0 in zip(chroms, start0s, end0s)
    )

    if as_array:
        return np.fromiter(val_gen, float)
    else:
        return list(val_gen)


def calculate_gc_fractions(chroms, start0s, end0s, fasta, window=None, as_array=True):
    logutils.log('Beginning gc fraction calculation')
    chrom_seq_cache = dict()
    get_fetchargs = _make_get_fetchargs(window, fasta)

    def gcfrac_generator():
        for chrom, start0, end0 in zip(chroms, start0s, end0s):
            chrom, start0, end0 = get_fetchargs(chrom, start0, end0)
            if chrom not in chrom_seq_cache:
                chrom_seq_cache[chrom] = fasta.fetch(chrom)
            seq = chrom_seq_cache[chrom][start0:end0]
            yield float(Bio.SeqUtils.gc_fraction(seq))

    if as_array:
        result = np.fromiter(gcfrac_generator(), float)
    else:
        result = list(gcfrac_generator())

    logutils.log('Finished gc fraction calculation')
    return result


def _make_get_fetchargs(window, fasta):
    if window is None:
        def get_fetchargs(chrom, start0, end0):
            return chrom, start0, end0
    else:
        chromlens = dict(zip(fasta.references, fasta.lengths))
        pad_left = int(window / 2)
        if window % 2 == 0:
            pad_right = pad_left
        else:
            pad_right = pad_left + 1

        def get_fetchargs(chrom, start0, end0):
            if end0 - start0 >= window:
                return chrom, start0, end0
            else:
                mid = int((start0 + end0 - 1) / 2)
                new_start0 = max(0, mid - pad_left)
                new_end0 = min(chromlens[chrom], mid + pad_right)
                return chrom, new_start0, new_end0

    return get_fetchargs



